# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch, Philipp Schubert
import datetime
import inspect
import logging
from math import nan
from pathlib import Path
from typing import Dict, List, Tuple, Union

import IPython
import numpy as np
import torch
import torch.utils.data
from tqdm import tqdm

from elektronn3.training import Trainer
from elektronn3.training import metrics
from elektronn3.training.train_utils import Timer
from elektronn3.training.trainer import NaNException

logger = logging.getLogger('elektronn3log')


class TrainerMulti(Trainer):
    """Experimental ``Trainer`` variant for specialized multilabel training.

    Not intended for general use. May move in the future."""

    def __init__(self, optimizer_iterations=1, loss_crop=16, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optimizer_iterations = optimizer_iterations
        assert(optimizer_iterations > 0)
        self.loss_crop = loss_crop  # 16 # crop sample for loss calcuation by this amount

    def run(self, max_steps: int = 1, max_runtime=3600 * 24 * 7) -> None:
        """Train the network for ``max_steps`` steps.

        After each training epoch, validation performance is measured and
        visualizations are computed and logged to tensorboard."""
        self.start_time = datetime.datetime.now()
        self.end_time = self.start_time + datetime.timedelta(seconds=max_runtime)
        self._save_model(suffix='_initial', verbose=False)
        self._lr_nhood.clear()
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])  # LR of the first training step
        while not self.terminate:
            try:
                stats, file_stats, misc, tr_sample_images = self._train(max_steps, max_runtime)
                self.epoch += 1

                if self.valid_dataset is None:
                    stats['val_loss'] = nan
                    val_sample_images = None
                else:
                    valid_stats, val_sample_images = self._validate()
                    stats.update(valid_stats)

                # Log to stdout and text log file
                self._log_basic(stats, misc)
                # Render visualizations and log to tensorboard
                self._log_to_tensorboard(stats, misc, tr_sample_images, val_sample_images, file_stats=file_stats)
                # Legacy non-tensorboard logging to files
                self._log_to_history_tracker(stats, misc)

                # Save trained model state
                self._save_model(val_loss=stats['val_loss'], verbose=False)  # Not verbose because it can get spammy.
                # TODO: Support other metrics for determining what's the "best" model?
                if stats['val_loss'] < self.best_val_loss:
                    self.best_val_loss = stats['val_loss']
                    self._save_model(suffix=f'_best{self.step}', val_loss=stats['val_loss'])
            except KeyboardInterrupt:
                if self.ipython_shell:
                    IPython.embed(header=self._shell_info)
                else:
                    break
                if self.terminate:
                    break
            except Exception as e:
                logger.exception('Unhandled exception during training:')
                if self.ignore_errors:
                    # Just print the traceback and try to carry on with training.
                    # This can go wrong in unexpected ways, so don't leave the training unattended.
                    pass
                elif self.ipython_shell:
                    print("\nEntering Command line such that Exception can be "
                          "further inspected by user.\n\n")
                    IPython.embed(header=self._shell_info)
                    if self.terminate:
                        break
                else:
                    raise e
        self._save_model(suffix='_final')

    def _train(self, max_steps, max_runtime):

        out_channels = self.out_channels

        def _channel_metric(metric, c, out_channels=out_channels, mean=False):
            """Returns an evaluator that calculates the ``metric``
            and selects its value for channel ``c``."""

            def evaluator(target, out):
                #pred = metrics._argmax(out)
                m = metric(target, out, num_classes=out_channels, ignore=out_channels - 1, mean=mean)
                return m[c]

            return evaluator

        tr_evaluators = {**{
            f'tr_DSC_c{c}': _channel_metric(metrics.dice_coefficient, c=c) for c in range(out_channels)
        }, **{
            f'tr_precision_c{c}': _channel_metric(metrics.precision, c=c) for c in range(out_channels)
        }, **{
            f'tr_recall_c{c}': _channel_metric(metrics.precision, c=c) for c in range(out_channels)
        }}
        # Scalar training stats that should be logged and written to tensorboard later
        stats: Dict[str, Union[float, List[float]]] = {stat: [] for stat in ['tr_loss', 'tr_loss_mean', 'tr_accuracy']}
        stats.update({name: [] for name in tr_evaluators.keys()})
        file_stats = {}
        # Other scalars to be logged
        misc: Dict[str, Union[float, List[float]]] = {misc: [] for misc in ['mean_target']}
        # Hold image tensors for real-time training sample visualization in tensorboard
        images: Dict[str, np.ndarray] = {}

        self.model.train()
        self.optimizer.zero_grad()
        running_vx_size = 0  # Counts input sizes (number of pixels/voxels) of training batches
        timer = Timer()
        import gc
        gc.collect()
        batch_iter = tqdm(self.train_loader, 'Training', total=len(self.train_loader))
        for i, batch in enumerate(batch_iter):
            if self.step in self.extra_save_steps:
                self._save_model(f'_step{self.step}', verbose=True)
            # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
            inp, target = batch['inp'], batch['target']
            cube_meta = batch['cube_meta']
            fname = batch['fname']
            dinp = inp.to(self.device, non_blocking=True)
            dtarget = target[:,:,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop].to(self.device, non_blocking=True) if self.loss_crop else target.to(self.device, non_blocking=True)
            weight = cube_meta[0].to(device=self.device, dtype=self.criterion.weight.dtype, non_blocking=True)
            prev_weight = self.criterion.weight.clone()
            self.criterion.weight = weight

            if isinstance(self.criterion, torch.nn.BCEWithLogitsLoss):
                ignore_mask = (1 - dtarget[0][-1]).view(1,1,*dtarget.shape[2:])
                dense_weight = self.criterion.weight.view(1,-1,1,1,1)
                positive_target_mask = (weight.view(1,-1,1,1,1) * dtarget)[0][1:-1].sum(dim=0).view(1,1,*dtarget.shape[2:]) # weighted targets w\ background and ignore
                needs_positive_target_mark = (dense_weight.sum() == 0).type(positive_target_mask.dtype)
                self.criterion.weight = ignore_mask * dense_weight + needs_positive_target_mark * positive_target_mask * prev_weight.view(1,-1,1,1,1)

            # forward pass
            dout = self.model(dinp)[:,:,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop] if self.loss_crop else self.model(dinp)

            #print(dout.dtype, dout.shape, dtarget.dtype, dtarget.shape, dout.min(), dout.max())
            dloss = self.criterion(dout, dtarget)
            #dcumloss = dloss if i == 0 else dcumloss + dloss
            #print(dloss, dloss.size())
            #dloss = (dloss * prev_weight * weight).mean()
            if torch.isnan(dloss).sum():
                logger.error('NaN loss detected! Aborting training.')
                raise NaNException

            if self.mixed_precision:
                from apex import amp
                with amp.scale_loss(dloss, self.optimizer) as scaled_loss:
                    scaled_loss.backward()
            else:
                # update step
                dloss.backward()

            if i % self.optimizer_iterations == self.optimizer_iterations - 1:
                self.optimizer.step()
                # TODO (lp): calling zero_grad() here makes gradients disappear from tb histograms
                self.optimizer.zero_grad()
                #loss2 = float(self.criterion(self.model(dinp), dtarget))
                #print(f'loss gain factor {np.divide(float(dloss), (float(dloss)-loss2))})')
            # End of core training loop on self.device

            with torch.no_grad():
                loss = float(dloss)
                # TODO: Evaluate performance impact of these copies and maybe avoid doing these so often
                out_class = dout.argmax(dim=1).detach().cpu()
                multi_class_target = target.argmax(1) if len(target.shape) > 4 else target  # TODO
                if self.loss_crop:
                    multi_class_target = multi_class_target[:,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop]
                acc = metrics.accuracy(multi_class_target, out_class, out_channels, mean=False).numpy()
                acc = np.average(acc[~np.isnan(acc)])#, weights=)
                mean_target = float(multi_class_target.to(torch.float32).mean())

                # import h5py
                # dsc5 = channel_metric(metrics.dice_coefficient, c=5, out_channels=out_channels)(multi_class_target, out_class)
                # after_step = '+' if i % self.optimizer_iterations == 0 else ''
                # with h5py.File(os.path.join(self.save_path, f'batch {self.step}{after_step} loss={float(dloss)} dsc5={dsc5}.h5'), "w") as f:
                #     f.create_dataset('raw', data=inp.squeeze(dim=0), compression="gzip")
                #     f.create_dataset('labels', data=multi_class_target.numpy().astype(np.uint16), compression="gzip")
                #     f.create_dataset('pred', data=dout.squeeze(dim=0).detach().cpu().numpy(), compression="gzip")

                if fname[0] not in file_stats:
                    file_stats[fname[0]] = []
                file_stats[fname[0]] += [float('nan')] * (i - len(file_stats[fname[0]])) + [loss]

                stats['tr_loss'].append(loss)
                stats['tr_loss_mean'] += [float('nan')] * (i - len(stats['tr_loss_mean']))
                if i % self.optimizer_iterations == self.optimizer_iterations - 1:
                    stats['tr_loss_mean'] += [np.mean(stats['tr_loss'][-self.optimizer_iterations:])]
                stats['tr_accuracy'].append(acc)
                for name, evaluator in tr_evaluators.items():
                    stats[name].append(evaluator(multi_class_target, out_class))

                misc['mean_target'].append(mean_target)
                # if loss-loss2 == 0 and not torch.any(out_class != multi_class_target):
                #     print('grad', self.model.up_convs[0].conv2.weight.grad)
                #     IPython.embed()
                #if loss - 0.99 < 1e-3:
                #    print('asd', loss, loss2)
                #    IPython.embed()
                batch_iter.set_description(f'Training (loss {loss:.4f})')
                #pbar.set_description(f'Training (loss {loss} / {float(dcumloss)})')
                #pbar.set_description(f'Training (loss {loss} / {np.divide(loss, (loss-loss2))})')
                self._tracker.update_timeline([self._timer.t_passed, loss, mean_target])

            self.criterion.weight = prev_weight

            # Not using .get_lr()[-1] because ReduceLROnPlateau does not implement get_lr()
            misc['learning_rate'] = self.optimizer.param_groups[0]['lr']  # LR for the this iteration
            # update schedules
            for sched in self.schedulers.values():
                # support ReduceLROnPlateau; doc. uses validation loss instead
                # http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau
                if "metrics" in inspect.signature(sched.step).parameters:
                    sched.step(metrics=loss)
                else:
                    sched.step()
            # Append LR of the next iteration (after sched.step()) for local LR minima detection
            self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])
            self._handle_lr()

            running_vx_size += inp.numel()

            #if stats['tr_loss_mean'][-1] < self.best_tr_loss:
            #   self.best_tr_loss = stats['tr_loss'][-1]
            #   self._save_model(suffix='_best_train', loss=stats['tr_loss'][-1])

            self.step += 1
            if self.step >= max_steps:
                logger.info(f'max_steps ({max_steps}) exceeded. Terminating...')
                self.terminate = True
            if datetime.datetime.now() >= self.end_time:
                logger.info(f'max_runtime ({max_runtime} seconds) exceeded. Terminating...')
                self.terminate = True
            if i == len(self.train_loader) - 1 or self.terminate:
                # Last step in this epoch or in the whole training
                # Preserve last training batch and network output for later visualization
                images['fname'] = Path(fname[0]).stem
                images['inp'] = inp.numpy()
                images['target'] = multi_class_target.numpy()
                images['out'] = dout.detach().cpu().numpy()
                self._put_current_attention_maps_into(images)

            if self.terminate:
                break

        stats['tr_loss_std'] = np.std(stats['tr_loss'])
        misc['tr_speed'] = len(self.train_loader) / timer.t_passed
        misc['tr_speed_vx'] = running_vx_size / timer.t_passed / 1e6  # MVx

        return stats, file_stats, misc, images

    @torch.no_grad()
    def _validate(self) -> Tuple[Dict[str, float], Dict[str, np.ndarray]]:
        self.model.eval()  # Set dropout and batchnorm to eval mode

        val_loss = []
        stats = {name: [] for name in self.valid_metrics.keys()}
        batch_iter = tqdm(self.valid_loader, 'Validating', total=len(self.valid_loader))
        for i, batch in enumerate(batch_iter):
            # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
            inp, target = batch['inp'], batch['target']
            cube_meta = batch['cube_meta']
            dinp = inp.to(self.device, non_blocking=True)
            dtarget = target[:,:,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop].to(self.device, non_blocking=True) if self.loss_crop else target.to(self.device, non_blocking=True)
            weight = cube_meta[0].to(device=self.device, dtype=self.criterion.weight.dtype, non_blocking=True)
            prev_weight = self.criterion.weight.clone()
            self.criterion.weight *= weight
            #self.criterion.pos_weight = self.criterion.weight

            if isinstance(self.criterion, torch.nn.BCEWithLogitsLoss):
                ignore_mask = (1 - dtarget[0][-1]).view(1,1,*dtarget.shape[2:])
                dense_weight = self.criterion.weight.view(1,-1,1,1,1)
                positive_target_mask = (weight.view(1,-1,1,1,1) * dtarget)[0][1:-1].sum(dim=0).view(1,1,*dtarget.shape[2:]) # weighted targets w\ background and ignore
                needs_positive_target_mark = (dense_weight.sum() == 0).type(positive_target_mask.dtype)
                self.criterion.weight = ignore_mask * dense_weight + needs_positive_target_mark * positive_target_mask * prev_weight.view(1,-1,1,1,1)

            dout = self.model(dinp)[:,:,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop] if self.loss_crop else self.model(dinp)
            multi_class_target = target.argmax(1) if len(target.shape) > 4 else target
            if self.loss_crop:
                multi_class_target = multi_class_target[:,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop,self.loss_crop:-self.loss_crop]
            val_loss.append(self.criterion(dout, dtarget).item())
            out = dout.detach().cpu()
            out_class = out.argmax(dim=1)
            self.criterion.weight = prev_weight
            for name, evaluator in self.valid_metrics.items():
                stats[name].append(evaluator(multi_class_target, out_class))
        images = {
            'fname': Path(batch['fname'][0]).stem,
            'inp': inp.numpy(),
            'out': out.numpy(),
            'target': multi_class_target.numpy()
        }
        self._put_current_attention_maps_into(images)

        stats['val_loss'] = np.mean(val_loss)
        stats['val_loss_std'] = np.std(val_loss)
        for name in self.valid_metrics.keys():
            stats[name] = np.nanmean(stats[name])

        self.model.train()  # Reset model to training mode

        return stats, images



from typing import Dict, Any, Tuple, Union, List

import numpy as np
import torch
from tqdm import tqdm

from elektronn3.training import Trainer, handlers
from elektronn3.training.train_utils import Timer
from elektronn3.training.trainer import logger, NaNException


class TripletTrainer(Trainer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def noop(*args, **kwargs): pass

        self.preview_plotting_handler = noop  # TODO
        self.sample_plotting_handler = handlers._tb_log_sample_images_all_img

    def _train_step_triplet(self, batch: Dict[str, Any]) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Core training step for triplet loss on self.device"""
        # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
        danchor = batch['anchor'].to(self.device, non_blocking=True)
        dpos = batch['pos'].to(self.device, non_blocking=True)
        dneg = batch['neg'].to(self.device, non_blocking=True)
        # forward pass
        danc_out = self.model(danchor)
        dpos_out = self.model(dpos)
        dneg_out = self.model(dneg)
        dloss = self.criterion(danc_out, dpos_out, dneg_out)
        if torch.isnan(dloss):
            logger.error('NaN loss detected! Aborting training.')
            raise NaNException
        # update step
        self.optimizer.zero_grad()
        dloss.backward()
        self.optimizer.step()
        return dloss, {'anchor_out': danc_out, 'pos_out': dpos_out, 'neg_out': dneg_out}

    def _train(self, max_steps, max_runtime):
        """Train for one epoch or until max_steps or max_runtime is reached"""
        self.model.train()

        # Scalar training stats that should be logged and written to tensorboard later
        stats: Dict[str, Union[float, List[float]]] = {stat: [] for stat in ['tr_loss']}
        # Other scalars to be logged
        misc: Dict[str, Union[float, List[float]]] = {misc: [] for misc in ['mean_target']}
        # Hold image tensors for real-time training sample visualization in tensorboard
        images: Dict[str, np.ndarray] = {}

        running_vx_size = 0  # Counts input sizes (number of pixels/voxels) of training batches
        timer = Timer()
        batch_iter = tqdm(
            self.train_loader, 'Training', total=len(self.train_loader), dynamic_ncols=True, **self.tqdm_kwargs
        )
        for i, batch in enumerate(batch_iter):
            if self.step in self.extra_save_steps:
                self._save_model(f'_step{self.step}', verbose=True)

            dloss, dout_imgs = self._train_step_triplet(batch)

            with torch.no_grad():
                loss = float(dloss)
                mean_target = 0.  # Dummy value
                misc['mean_target'].append(mean_target)
                stats['tr_loss'].append(loss)
                batch_iter.set_description(f'Training (loss {loss:.4f})')
                self._tracker.update_timeline([self._timer.t_passed, loss, mean_target])

            # Not using .get_lr()[-1] because ReduceLROnPlateau does not implement get_lr()
            misc['learning_rate'] = self.optimizer.param_groups[0]['lr']  # LR for the this iteration
            self._scheduler_step(loss)

            running_vx_size += batch['anchor'].numel()

            self._incr_step(max_runtime, max_steps)
            if i == len(self.train_loader) - 1 or self.terminate:
                # Last step in this epoch or in the whole training
                # Preserve last training batch and network output for later visualization
                for key, img in batch.items():
                    if isinstance(img, torch.Tensor):
                        img = img.detach().cpu().numpy()
                    images[key] = img
                self._put_current_attention_maps_into(images)

                # TODO: The plotting handler abstraction is inadequate here. Figure out how
                #       we can handle plotting cleanly in one place.
                # Outputs are visualized here, while inputs are visualized in the plotting handler
                #  which is called in _run()...
                for name, img in dout_imgs.items():
                    img = img.detach()[0].cpu().numpy()  # select first item of batch
                    for c in range(img.shape[0]):
                        if img.ndim == 4:  # 3D data
                            img = img[:, img.shape[0] // 2]  # take center slice of depth dim -> 2D
                        self.tb.add_figure(
                            f'tr_samples/{name}_c{c}',
                            handlers.plot_image(img[c], cmap='gray'),
                            global_step=self.step
                        )

            if self.terminate:
                break

        stats['tr_loss_std'] = np.std(stats['tr_loss'])
        misc['tr_speed'] = len(self.train_loader) / timer.t_passed
        misc['tr_speed_vx'] = running_vx_size / timer.t_passed / 1e6  # MVx

        return stats, misc, images

    def _validate(self) -> Tuple[Dict[str, float], Dict[str, np.ndarray]]:
        raise NotImplementedError



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany

"""
adopted from https://github.com/bfortuner/pytorch_tiramisu/blob/master/models/
LICENSE https://github.com/bfortuner/pytorch_tiramisu/blob/master/LICENSE
"""


import torch
import torch.nn as nn


class FCDenseNet(nn.Module):
    def __init__(self, in_channels=3, down_blocks=(5,5,5,5,5),
                 up_blocks=(5,5,5,5,5), bottleneck_layers=5,
                 growth_rate=16, out_chans_first_conv=48, n_classes=12):
        super().__init__()
        self.down_blocks = down_blocks
        self.up_blocks = up_blocks
        cur_channels_count = 0
        skip_connection_channel_counts = []

        ## First Convolution ##

        self.add_module('firstconv', nn.Conv2d(in_channels=in_channels,
                  out_channels=out_chans_first_conv, kernel_size=3,
                  stride=1, padding=1, bias=True))
        cur_channels_count = out_chans_first_conv

        #####################
        # Downsampling path #
        #####################

        self.denseBlocksDown = nn.ModuleList([])
        self.transDownBlocks = nn.ModuleList([])
        for i in range(len(down_blocks)):
            self.denseBlocksDown.append(
                DenseBlock(cur_channels_count, growth_rate, down_blocks[i]))
            cur_channels_count += (growth_rate*down_blocks[i])
            skip_connection_channel_counts.insert(0,cur_channels_count)
            self.transDownBlocks.append(TransitionDown(cur_channels_count))

        #####################
        #     Bottleneck    #
        #####################

        self.add_module('bottleneck',Bottleneck(cur_channels_count,
                                     growth_rate, bottleneck_layers))
        prev_block_channels = growth_rate*bottleneck_layers
        cur_channels_count += prev_block_channels

        #######################
        #   Upsampling path   #
        #######################

        self.transUpBlocks = nn.ModuleList([])
        self.denseBlocksUp = nn.ModuleList([])
        for i in range(len(up_blocks)-1):
            self.transUpBlocks.append(TransitionUp(prev_block_channels, prev_block_channels))
            cur_channels_count = prev_block_channels + skip_connection_channel_counts[i]

            self.denseBlocksUp.append(DenseBlock(
                cur_channels_count, growth_rate, up_blocks[i],
                    upsample=True))
            prev_block_channels = growth_rate*up_blocks[i]
            cur_channels_count += prev_block_channels

        ## Final DenseBlock ##

        self.transUpBlocks.append(TransitionUp(
            prev_block_channels, prev_block_channels))
        cur_channels_count = prev_block_channels + skip_connection_channel_counts[-1]

        self.denseBlocksUp.append(DenseBlock(
            cur_channels_count, growth_rate, up_blocks[-1],
                upsample=False))
        cur_channels_count += growth_rate*up_blocks[-1]

        ## Softmax ##

        self.finalConv = nn.Conv2d(in_channels=cur_channels_count,
               out_channels=n_classes, kernel_size=1, stride=1,
                   padding=0, bias=True)
        self.softmax = nn.LogSoftmax(dim=1)

    def forward(self, x):
        out = self.firstconv(x)

        skip_connections = []
        for i in range(len(self.down_blocks)):
            out = self.denseBlocksDown[i](out)
            skip_connections.append(out)
            out = self.transDownBlocks[i](out)

        out = self.bottleneck(out)
        for i in range(len(self.up_blocks)):
            skip = skip_connections.pop()
            out = self.transUpBlocks[i](out, skip)
            out = self.denseBlocksUp[i](out)

        out = self.finalConv(out)
        out = self.softmax(out)
        return out


def FCDenseNet57(n_classes):
    return FCDenseNet(
        in_channels=3, down_blocks=(4, 4, 4, 4, 4),
        up_blocks=(4, 4, 4, 4, 4), bottleneck_layers=4,
        growth_rate=12, out_chans_first_conv=48, n_classes=n_classes)


def FCDenseNet67(n_classes):
    return FCDenseNet(
        in_channels=3, down_blocks=(5, 5, 5, 5, 5),
        up_blocks=(5, 5, 5, 5, 5), bottleneck_layers=5,
        growth_rate=16, out_chans_first_conv=48, n_classes=n_classes)


def FCDenseNet103(n_classes):
    return FCDenseNet(
        in_channels=3, down_blocks=(4,5,7,10,12),
        up_blocks=(12,10,7,5,4), bottleneck_layers=15,
        growth_rate=16, out_chans_first_conv=48, n_classes=n_classes)



class DenseLayer(nn.Sequential):
    def __init__(self, in_channels, growth_rate):
        super().__init__()
        self.add_module('norm', nn.BatchNorm2d(in_channels))
        self.add_module('relu', nn.ReLU(True))
        self.add_module('conv', nn.Conv2d(in_channels, growth_rate, kernel_size=3,
                                          stride=1, padding=1, bias=True))
        self.add_module('drop', nn.Dropout2d(0.2))

    def forward(self, x):
        return super().forward(x)


class DenseBlock(nn.Module):
    def __init__(self, in_channels, growth_rate, n_layers, upsample=False):
        super().__init__()
        self.upsample = upsample
        self.layers = nn.ModuleList([DenseLayer(
            in_channels + i*growth_rate, growth_rate)
            for i in range(n_layers)])

    def forward(self, x):
        if self.upsample:
            new_features = []
            #we pass all previous activations into each dense layer normally
            #But we only store each dense layer's output in the new_features array
            for layer in self.layers:
                out = layer(x)
                x = torch.cat([x, out], 1)
                new_features.append(out)
            return torch.cat(new_features,1)
        else:
            for layer in self.layers:
                out = layer(x)
                x = torch.cat([x, out], 1) # 1 = channel axis
            return x


class TransitionDown(nn.Sequential):
    def __init__(self, in_channels):
        super().__init__()
        self.add_module('norm', nn.BatchNorm2d(num_features=in_channels))
        self.add_module('relu', nn.ReLU(inplace=True))
        self.add_module('conv', nn.Conv2d(in_channels, in_channels,
                                          kernel_size=1, stride=1,
                                          padding=0, bias=True))
        self.add_module('drop', nn.Dropout2d(0.2))
        self.add_module('maxpool', nn.MaxPool2d(2))

    def forward(self, x):
        return super().forward(x)


class TransitionUp(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.convTrans = nn.ConvTranspose2d(
            in_channels=in_channels, out_channels=out_channels,
            kernel_size=3, stride=2, padding=0, bias=True)

    def forward(self, x, skip):
        out = self.convTrans(x)
        out = center_crop(out, skip.size(2), skip.size(3))
        out = torch.cat([out, skip], 1)
        return out


class Bottleneck(nn.Sequential):
    def __init__(self, in_channels, growth_rate, n_layers):
        super().__init__()
        self.add_module('bottleneck', DenseBlock(
            in_channels, growth_rate, n_layers, upsample=True))

    def forward(self, x):
        return super().forward(x)


def center_crop(layer, max_height, max_width):
    _, _, h, w = layer.size()
    xy1 = (w - max_width) // 2
    xy2 = (h - max_height) // 2
    return layer[:, :, xy2:(xy2 + max_height), xy1:(xy1 + max_width)]


# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch

# TODO: Update docs to show Evaluator
"""Metrics and tools for evaluating neural network predictions

References:

- https://en.wikipedia.org/wiki/Confusion_matrix
- https://stats.stackexchange.com/questions/273537/f1-dice-score-vs-iou
- http://scikit-learn.org/stable/modules/model_evaluation.html

.. note::

    ``sklearn.metrics`` has a lot of alternative implementations that can be
    compared with these here and could be used as inspiration for future work
    (http://scikit-learn.org/stable/modules/classes.html#classification-metrics).

    For example, to get the equivalent output to
    ``elektronn3.training.metrics.recall(target, pred, num_classes=2, mean=False) / 100``,
    from scikit-learn, you can compute
    ``sklearn.metrics.recall_score(target.view(-1).cpu().numpy(), pred.view(-1).cpu().numpy(), average=None).astype(np.float32)``.


    For most metrics, we don't use scikit-learn directly in this module for
    performance reasons:

    - PyTorch allows us to calculate metrics directly on GPU
    - We LRU-cache confusion matrices for cheap calculation of multiple metrics
"""

from functools import lru_cache
from typing import Callable, Optional
from unicodedata import name

import sklearn.metrics
import torch
import numpy as np


eps = 0.0001  # To avoid divisions by zero

# TODO: Tests would make a lot of sense here.


@lru_cache(maxsize=128)
def confusion_matrix(
        target: torch.LongTensor,
        pred: torch.LongTensor,
        num_classes: int = 2,
        dtype: torch.dtype = torch.float32,
        device: torch.device = torch.device('cpu'),
        nan_when_empty: bool = True,
        ignore: Optional[int] = None,
) -> torch.Tensor:
    """ Calculate per-class confusion matrix.

    Uses an LRU cache, so subsequent calls with the same arguments are very
    cheap.

    Args:
        pred: Tensor with predicted class values
        target: Ground truth tensor with true class values
        num_classes: Number of classes that the target can assume.
            E.g. for binary classification, this is 2.
            Classes are expected to start at 0
        dtype: ``torch.dtype`` to be used for calculation and output.
            ``torch.float32`` is used as default because it is robust
            against overflows and can be used directly in true divisions
            without re-casting.
        device: PyTorch device on which to store the confusion matrix
        nan_when_empty: If ``True`` (default), the confusion matrix will
            be filled with NaN values for each channel of which there are
            no positive entries in the ``target`` tensor.
        ignore: Index to be ignored for cm calculation

    Returns:
        Confusion matrix ``cm``, with shape ``(num_classes, 4)``, where
        each row ``cm[c]`` contains (in this order) the count of
        - true positives
        - true negatives
        - false positives
        - false negatives
        of ``pred`` w.r.t. ``target`` and class ``c``.

        E.g. ``cm[1][2]`` contains the number of false positive predictions
        of class ``1``.
        If ``nan_when_empty`` is enabled and there are no positive elements
        of class ``1`` in ``target``, ``cm[1]`` will instead be filled with
        NaN values.
    """
    cm = torch.empty(num_classes, 4, dtype=dtype, device=device)
    for c in range(num_classes):
        pos_pred = pred == c
        neg_pred = ~pos_pred
        pos_target = target == c
        if ignore is not None:
            ign_target = target == ignore
        else:
            ign_target = False  # Makes `& ~ign_target` a no-op
        # Manual conversion to Tensor because of a type promotion regression in PyTorch 1.5
        ign_target = torch.tensor(ign_target, dtype=torch.bool, device=device)
        neg_target = ~pos_target

        true_pos = (pos_pred & pos_target & ~ign_target).sum(dtype=dtype)
        true_neg = (neg_pred & neg_target & ~ign_target).sum(dtype=dtype)
        false_pos = (pos_pred & neg_target & ~ign_target).sum(dtype=dtype)
        false_neg = (neg_pred & pos_target & ~ign_target).sum(dtype=dtype)

        cm[c] = torch.tensor([true_pos, true_neg, false_pos, false_neg])

        if nan_when_empty and pos_target.sum(dtype=dtype) == 0:
            cm[c] = torch.tensor([float('nan')] * 4)

    return cm


def precision(target, pred, num_classes=2, mean=True, ignore=None):
    """Precision metric (in %)"""
    cm = confusion_matrix(target, pred, num_classes=num_classes, ignore=ignore)
    tp, tn, fp, fn = cm.transpose(0, 1)  # Transposing to put class axis last
    # Compute metrics for each class simulataneously
    prec = tp / (tp + fp + eps)  # Per-class precision
    if mean:
        prec = prec.mean().item()
    return prec * 100


def recall(target, pred, num_classes=2, mean=True, ignore=None):
    """Recall metric a.k.a. sensitivity a.k.a. hit rate (in %)"""
    cm = confusion_matrix(target, pred, num_classes=num_classes, ignore=ignore)
    tp, tn, fp, fn = cm.transpose(0, 1)  # Transposing to put class axis last
    rec = tp / (tp + fn + eps)  # Per-class recall
    if mean:
        rec = rec.mean().item()
    return rec * 100


def accuracy(target, pred, num_classes=2, mean=True, ignore=None):
    """Accuracy metric (in %)"""
    cm = confusion_matrix(target, pred, num_classes=num_classes, ignore=ignore)
    tp, tn, fp, fn = cm.transpose(0, 1)  # Transposing to put class axis last
    acc = (tp + tn) / (tp + tn + fp + fn + eps)  # Per-class accuracy
    if mean:
        acc = acc.mean().item()
    return acc * 100


def dice_coefficient(target, pred, num_classes=2, mean=True, ignore=None):
    """Sørensen–Dice coefficient a.k.a. DSC a.k.a. F1 score (in %)"""
    cm = confusion_matrix(target, pred, num_classes=num_classes, ignore=ignore)
    tp, tn, fp, fn = cm.transpose(0, 1)  # Transposing to put class axis last
    dsc = 2 * tp / (2 * tp + fp + fn + eps)  # Per-class (Sørensen-)Dice similarity coefficient
    if mean:
        dsc = dsc.mean().item()
    return dsc * 100


def iou(target, pred, num_classes=2, mean=True, ignore=None):
    """IoU (Intersection over Union) a.k.a. IU a.k.a. Jaccard index (in %)"""
    cm = confusion_matrix(target, pred, num_classes=num_classes, ignore=ignore)
    tp, tn, fp, fn = cm.transpose(0, 1)  # Transposing to put class axis last
    iu = tp / (tp + fp + fn + eps)  # Per-class Intersection over Union
    if mean:
        iu = iu.mean().item()
    return iu * 100


def auroc(target, probs, mean=True):
    """ Area under Curve (AuC) of the ROC curve (in %).

    .. note::
        This implementation uses scikit-learn on the CPU to do the heavy
        lifting, so it's relatively slow (one call can take about 1 second
        for typical inputs).
    """
    assert probs.dim() == target.dim() + 1
    num_classes = probs.shape[1]
    # target: (N, [D,], H, W) -> (N*[D,]*H*W,)
    target_npflat = target.view(-1).cpu().numpy()
    # probs: (N, C, [D,], H, W) -> (C, N*[D,]*H*W)
    probs_npflat = probs.transpose(1, 0).view(num_classes, -1).cpu().numpy()
    auc = torch.empty(num_classes)
    # Direct roc_auc_score() computation with multi-class arrays can take
    #  hours, so split this into binary calculations manually here by looping
    #  through classes:
    for c in range(num_classes):
        t = target_npflat == c  # 1 where target is c, 0 everywhere else
        p = probs_npflat[c]  # probs of class c
        auc[c] = sklearn.metrics.roc_auc_score(t, p)
    if mean:
        auc = auc.mean().item()
    return auc * 100


def average_precision(target, probs, mean=True):
    """Average precision (AP) metric based on PR curves (in %).

    .. note::
        This implementation uses scikit-learn on the CPU to do the heavy
        lifting, so it's relatively slow (one call can take about 1 second
        for typical inputs).
    """
    assert probs.dim() == target.dim() + 1
    num_classes = probs.shape[1]
    # target: (N, [D,], H, W) -> (N*[D,]*H*W,)
    target_npflat = target.view(-1).cpu().numpy()
    # probs: (N, C, [D,], H, W) -> (C, N*[D,]*H*W)
    probs_npflat = probs.transpose(1, 0).view(num_classes, -1).cpu().numpy()
    ap = torch.empty(num_classes)
    # Direct average_precision_score() computation with multi-class arrays can take
    #  hours, so split this into binary calculations manually here by looping
    #  through classes:
    for c in range(num_classes):
        t = target_npflat == c  # 1 where target is c, 0 everywhere else
        p = probs_npflat[c]  # probs of class c
        ap[c] = sklearn.metrics.average_precision_score(t, p)
    if mean:
        ap = ap.mean().item()
    return ap * 100


@lru_cache(maxsize=128)
def _softmax(x, dim=1):
    return torch.nn.functional.softmax(x, dim)


@lru_cache(maxsize=128)
def _argmax(x, dim=1):
    return x.argmax(dim)


# Helper for multi-class metric construction
def channel_metric(metric, c, num_classes, argmax=True):
    """Returns an evaluator that calculates the ``metric``
    and selects its value for channel ``c``.

    Example:
        >>> from elektronn3.training import metrics
        >>> num_classes = 5  # Example. Depends on model and data set
        >>> # Metric evaluator dict that registers DSCs of all output channels.
        >>> # You can pass it to elektronn3.training.Trainer as the ``valid_metrics``
        >>> #  argument to make it log these values.
        >>> dsc_evaluators = {
        ...    f'val_DSC_c{c}': channel_metric(
        ...        metrics.dice_coefficient,
        ...        c=c, num_classes=num_classes
        ...    )
        ...    for c in range(num_classes)
        ... }
    """
    def evaluator(target, out):
        pred = _argmax(out) if argmax else out
        m = metric(target, pred, num_classes=num_classes, mean=False)
        return m[c]

    return evaluator


# Metric evaluator shortcuts for raw network outputs in binary classification
#  tasks ("bin_*"). "Raw" means not softmaxed or argmaxed.

# These are deprecated and will be removed later. Use Evaluators instead.

def bin_precision(target, out):
    pred = _argmax(out)
    return precision(
        target, pred, num_classes=2, mean=False
    )[1]  # Take only the score for class 1


def bin_recall(target, out):
    pred = _argmax(out)
    return recall(
        target, pred, num_classes=2, mean=False
    )[1]  # Take only the score for class 1


def bin_accuracy(target, out):
    pred = _argmax(out)
    return accuracy(
        target, pred, num_classes=2, mean=False
    )[1]  # Take only the score for class 1


def bin_dice_coefficient(target, out):
    pred = _argmax(out)
    return dice_coefficient(
        target, pred, num_classes=2, mean=False
    )[1]  # Take only the score for class 1


def bin_iou(target, out):
    pred = _argmax(out)
    return iou(
        target, pred, num_classes=2, mean=False
    )[1]  # Take only the score for class 1


def bin_average_precision(target, out):
    probs = _softmax(out)
    return average_precision(
        target, probs, mean=False
    )[1]  # Take only the score for class 1


def bin_auroc(target, out):
    probs = _softmax(out)
    return auroc(
        target, probs, mean=False
    )[1]  # Take only the score for class 1


class Evaluator:
    name: str = 'generic'

    def __init__(
            self,
            metric_fn: Callable,
            index: Optional[int] = None,
            ignore: Optional[int] = None,
            self_supervised: Optional[bool] = False
    ):
        self.metric_fn = metric_fn
        self.index = index
        self.ignore = ignore
        self.num_classes = None
        self.self_supervised = self_supervised

    def __call__(self, target: torch.Tensor, out: torch.Tensor) -> float:
        # self supervised training 
        if self.self_supervised:
            m = self.metric_fn(target, out)
            return m  # aggregation handled by sklearn metric
        # supervised training
        else:
            if self.num_classes is None:
                self.num_classes = out.shape[1]
            # print(self.num_classes)
            pred = _argmax(out)
            m = self.metric_fn(target, pred, self.num_classes, mean=False, ignore=self.ignore)
            if self.index is None:
                return m.mean().item()
            return m[self.index].item()


class Accuracy(Evaluator):
    name = 'accuracy'
    def __init__(self, *args, **kwargs): super().__init__(accuracy, *args, **kwargs)


class Precision(Evaluator):
    name = 'precision'
    def __init__(self, *args, **kwargs): super().__init__(precision, *args, **kwargs)


class Recall(Evaluator):
    name = 'recall'
    def __init__(self, *args, **kwargs): super().__init__(recall, *args, **kwargs)


class IoU(Evaluator):
    name = 'IoU'
    def __init__(self, *args, **kwargs): super().__init__(iou, *args, **kwargs)


class DSC(Evaluator):
    name = 'DSC'
    def __init__(self, *args, **kwargs): super().__init__(dice_coefficient, *args, **kwargs)


class AveragePrecision(Evaluator):
    name = 'AP'
    def __init__(self, *args, **kwargs): super().__init__(sklearn.metrics.average_precision_score, *args, **kwargs)


class AUROC(Evaluator):
    name = 'AUROC'
    def __init__(self, *args, **kwargs): super().__init__(sklearn.metrics.roc_auc_score, *args, **kwargs)


class NMI(Evaluator):
    name = 'NMI'
    def __init__(self, *args, **kwargs): super().__init__(sklearn.metrics.v_measure_score, *args, **kwargs)


class AMI(Evaluator):
    name = 'AMI'
    def __init__(self, *args, **kwargs): super().__init__(sklearn.metrics.adjusted_mutual_info_score, *args, **kwargs)


class SilhouetteScore(Evaluator):
    name = 'silhouette_score'
    def __init__(self, *args, **kwargs): super().__init__(sklearn.metrics.silhouette_score, *args, **kwargs)


class ARI(Evaluator):
    name = 'ARI'
    def __init__(self, *args, **kwargs): super().__init__(sklearn.metrics.adjusted_rand_score, *args, **kwargs)


__all__ = ["floatX"]

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

import numpy as np
from elektronn3.logger import logger_setup
import logging
logger = logging.getLogger('elektronn3log')

logger_setup()

floatX = np.float32  # TODO: Either hardcode float32 everywhere or add float16 support


def select_mpl_backend(mpl_backend):
    """ Set up matplotlib to use the specified backend.

    This needs to be run BEFORE the first import of matplotlib.pyplot!
    """
    import os
    from subprocess import check_call, CalledProcessError
    import matplotlib
    if mpl_backend.lower() == 'agg':
        matplotlib.use('AGG')
        logger.info('Using the AGG backend for matplotlib. No support for X11 windows.')
    else:
        if mpl_backend.startswith('force-'):
            matplotlib.use(mpl_backend.partition('force-')[-1])
        else:
            # Prevent setting of mpl qt-backend on machines without X-server before other modules import mpl.
            with open(os.devnull, 'w') as devnull:
                try:
                    # "xset q" will always succeed to run if an X server is currently running
                    check_call(['xset', 'q'], stdout=devnull, stderr=devnull)
                    if mpl_backend.lower() == 'auto':
                        pass  # Backend is silently set to system default.
                    else:
                        matplotlib.use(mpl_backend)
                    print('Using the {} backend for matplotlib.'.format(matplotlib.get_backend()))
                    # Don't set backend explicitly, use system default...
                # if "xset q" fails, conclude that X is not running
                except (OSError, CalledProcessError):
                    print('No X11 server found. Falling back to AGG backend for matplotlib.')
                    matplotlib.use('AGG')



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch, Philipp Schubert
import datetime
import pprint
from collections import deque

import gc
import logging
import os
import shutil
import warnings
import zipfile

from itertools import islice
from math import nan
from pickle import PickleError
from textwrap import dedent
from typing import Tuple, Dict, Optional, Callable, Any, Sequence, List, Union

import inspect
import IPython
import numpy as np
import tensorboardX
import torch
import torch.utils.data
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import StepLR
from torch.cuda import amp
from tqdm import tqdm


import elektronn3
from elektronn3.training import handlers
from elektronn3.training.swa import SWA
from elektronn3.training.train_utils import pretty_string_time, create_preview_batch_from_knossos
from elektronn3.training.train_utils import Timer, HistoryTracker

from torch.utils import collect_env
from elektronn3.inference import Predictor
from elektronn3 import __file__ as arch_src

logger = logging.getLogger('elektronn3log')


class NaNException(RuntimeError):
    """When a NaN value is detected"""
    pass


def _worker_init_fn(worker_id: int) -> None:
    """Sets a unique but deterministic random seed for background workers.

    Only sets the seed for NumPy because PyTorch and Python's own RNGs
    take care of reseeding on their own.
    See https://github.com/numpy/numpy/issues/9650."""
    # Modulo 2**32 because np.random.seed() only accepts values up to 2**32 - 1
    initial_seed = torch.initial_seed() % 2**32
    worker_seed = initial_seed + worker_id
    np.random.seed(worker_seed)


# Be careful from where you call this! Not sure if this is concurrency-safe.
def _change_log_file_to(
        new_path: str,
        transfer_old_logs: bool = True,
        delete_old_file: bool = True
) -> None:
    """Transfer the current log file to a new location and redirect logs."""

    def _get_first_file_handler() -> logging.FileHandler:
        for handler in logger.handlers:
            if isinstance(handler, logging.FileHandler):
                return handler
        return RuntimeError('logger has no FileHandler.')

    # Getting the first (and presumably only) file handler
    file_handler = _get_first_file_handler()
    if transfer_old_logs:
        with open(file_handler.baseFilename) as f:
            old_logs = f.read()
        with open(new_path, 'w') as f:
            f.write(old_logs)
    file_handler.close()
    if delete_old_file:
        os.remove(file_handler.baseFilename)
    file_handler.baseFilename = new_path


class Trainer:
    """ General training loop abstraction for supervised training.

    Args:
        model: PyTorch model (``nn.Module``) that shall be trained.
            Please make sure that the output shape of the ``model``
            matches the shape of targets that are delivered by the
            ``train_dataset``.
        criterion: PyTorch loss that shall be used as the optimization
            criterion.
        optimizer: PyTorch optimizer that shall be used to update
            ``model`` weights according to the ``criterion`` in each
            iteration.
        device: The device on which the network shall be trained.
        train_dataset: PyTorch dataset (``data.Dataset``) which produces
            training samples when iterated over.
            :py:class:`elektronn3.data.cnndata.PatchCreator` is currently
            recommended for constructing datasets.
        valid_dataset: PyTorch dataset (``data.Dataset``) which produces
            validation samples when iterated over.
            The length (``len(valid_dataset)``) of it determines how many
            samples are used for one validation metric calculation.
        unlabeled_dataset: Unlabeled dataset (only inputs) for
            semi-supervised training. If this is supplied, ``ss_criterion``
            needs to be set to the loss that should be computed on unlabeled
            inputs.
        valid_metrics: Validation metrics to be calculated on
            validation data after each training epoch. All metrics are logged
            to tensorboard.
        ss_criterion: Loss criterion for the self-supervised part of
            semi-supervised training. The ``ss_criterion`` loss is computed
            on batches from the ``unlabeled_dataset`` and added to the
            supervised loss in each training step.
        save_root: Root directory where training-related files are
            stored. Files are always written to the subdirectory
            ``save_root/exp_name/``.
        exp_name: Name of the training experiment. Determines the subdirectory
            to which files are written and should uniquely identify one
            training experiment.
            If ``exp_name`` is not set, it is auto-generated from the model
            name and a time stamp in the format ``'%y-%m-%d_%H-%M-%S'``.
        example_input: An example input tensor that can be fed to the
            ``model``. This is used for JIT tracing during model serialization.
        save_jit: Chooses if/how a JIT version (.pts file) of the
            ``model`` should always be saved in addition to regular model
            snapshots. Choices:

            - ``None`` (default): Disable saving JIT models.
            - ``'script'`` (recommended if possible): The model is compiled
              directly with ``torch.jit.script()`` and saved as a .pts file
            - ``'trace'``: The model is JIT-traced with ``example_input``
              and saved as a .pts file
        batch_size: Desired batch size of training samples.
        preview_batch: Set a fixed input batch for preview predictions.
            If it is ``None`` (default), preview batch functionality will be
            disabled. As a more powerful alternative for KNOSSOS datasets, consider using
            the ``knossos_preview_config`` option instead.
        knossos_preview_config: Configures preview batch creation and preview inferences
            based on a KNOSSOS dataset region. Here is an example of how it should look like:

            >>> knossos_preview_config = {
            ...     'dataset': 'path/to/knossos/dataset.conf',
            ...     'offset': [0, 0, 0],  # Offset (min) coordinates
            ...     'size': [256, 256, 64],  # Size (shape) of the region
            ...     'mag': 1,  # source mag
            ...     'target_mags': [1, 2, 3],  # List of target mags to which the inference results should be written
            ...     'remap_ids': None  # Config for class ID remapping (optional). See transforms.RemapTargetIDs
            ... }

            Periodic preview inference results are written to .k.zip annotation files that can be
            loaded with KNOSSOS and overlayed over the original data. .k.zip files are saved in the
            training directory, with file names reflecting their training step.
        preview_interval: Determines how often to perform preview inference.
            Preview inference is performed every ``preview_interval`` epochs
            during training. Regardless of this value, preview predictions
            will also be performed once after epoch 1.
            (To disable preview predictions altogether, just set
            ``preview_batch = None``).
        inference_kwargs: Additional options that are supplied to the
            :py:class:`elektronn3.inference.Predictor` instance
            that is used for periodic preview inference on the
            ``preview_batch``.
        extra_save_steps: Permanent model snapshots are saved at the
            training steps specified here. E.g. with
            ``extra_save_steps = (0, 30, 3000)``, a snapshot is made at
            steps 0 (before training begins), step 30 and step 3000.
        num_workers: Number of background processes that are used to produce
            training samples without blocking the main training loop.
            See :py:class:`torch.utils.data.DataLoader`
            For normal training, you can mostly set ``num_workers=1``.
            Only use more workers if you notice a data loader bottleneck.
            Set ``num_workers=0`` if you want to debug the datasets
            implementation, to avoid mulitprocessing-specific issues.
        schedulers: Dictionary of schedulers for training hyperparameters,
            e.g. learning rate schedulers that can be found in
            `py:mod:`torch.optim.lr_scheduler`.
        overlay_alpha: Alpha (transparency) value for alpha-blending of
            overlay image plots.
        enable_videos: Enables video visualizations for 3D image data
            in tensorboard. Requires the moviepy package.
            Warning: Videos are stored as GIFs and can get very large,
            so only use this if you log seldomly or have a lot of storage
            capacity.
        enable_tensorboard: If ``True``, tensorboard logging/plotting is
            enabled during training.
        tensorboard_root_path: Path to the root directory under which
            tensorboard log directories are created. Log ("event") files are
            written to a subdirectory that has the same name as the
            ``exp_name``.
            If ``tensorboard_root_path`` is not set, tensorboard logs are
            written to ``save_path`` (next to model checkpoints, plots etc.).
        ignore_errors: If ``True``, the training process tries to ignore
            all errors and continue with the next batch if it encounters
            an error on the current batch.
            It's not recommended to use this. It's only helpful for certain
            debugging scenarios.
        ipython_shell: If ``True`` keyboard interrupts (Ctrl-C) won't exit
            the process but only pause training and enter an IPython shell.
            Additionally, errors during training (except
            C-level segfaults etc.) won't crash the whole training process,
            but drop to an IPython shell so errors can be inspected with
            access to the current training state.
        out_channels: Optionally specifies the total number of different target
            classes for classification tasks. If this is not set manually,
            the ``Trainer`` checks if the ``train_dataset`` provides this
            value. If available, ``self.out_channels`` is set to
            ``self.train_dataset.out_channels``. Otherwise, it is set to
            ``None``.
            The ``out_channels`` attribute is used for plotting purposes and is
            not strictly required for training.
        sample_plotting_handler: Function that receives training and
            validation samples and is responsible for visualizing them by
            e.g. plotting them to tensorboard and/or writing them to files.
            It is called once after each training epoch and once after each
            validation run.
            If ``None``, a tensorboard-based default handler is used that
            works for most classification scenarios and for 3-channel
            regression.
        preview_plotting_handler: Function that is responsible for producing
            previews and visualizing/plotting/logging them.
            It is called once each ``preview_interval`` epochs.
            If ``None``, a tensorboard-based default handler is used that
            works for most classification scenarios.
        mixed_precision: If ``True``, enable Automated Mixed Precision training
            powered by https://github.com/NVIDIA/apex to reduce memory usage
            and (if a GPU with Tensor Cores is used) make training much faster.
            This is currently experimental and might cause instabilities.
        tqdm_kwargs: Extra arguments to be passed to tqdm progress bars.
            For example, to disable tqdm outputs completely, pass
            ``tqdm_kwargs={'disable': True}``.
    """

    tb: tensorboardX.SummaryWriter
    terminate: bool
    step: int
    epoch: int
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader
    exp_name: str
    save_path: str  # Full path to where training files are stored
    out_channels: Optional[int]  # Number of channels of the network outputs

    def __init__(
            self,
            model: torch.nn.Module,
            criterion: torch.nn.Module,
            optimizer: torch.optim.Optimizer,
            device: torch.device,
            save_root: str,
            train_dataset: torch.utils.data.Dataset,
            valid_dataset: Optional[torch.utils.data.Dataset] = None,
            unlabeled_dataset: Optional[torch.utils.data.Dataset] = None,
            valid_metrics: Optional[Dict] = None,
            ss_criterion: Optional[torch.nn.Module] = None,
            preview_batch: Optional[torch.Tensor] = None,
            knossos_preview_config: Optional[Dict[str, str]] = None,
            preview_interval: int = 5,
            inference_kwargs: Optional[Dict[str, Any]] = None,
            hparams: Optional[Dict[str, Any]] = None,
            extra_save_steps: Sequence[int] = (),
            exp_name: Optional[str] = None,
            example_input: Optional[torch.Tensor] = None,
            enable_save_trace: bool = False,
            save_jit: Optional[str] = None,
            batch_size: int = 1,
            num_workers: int = 0,
            schedulers: Optional[Dict[Any, Any]] = None,
            overlay_alpha: float = 0.4,
            enable_videos: bool = False,
            enable_tensorboard: bool = True,
            tensorboard_root_path: Optional[str] = None,
            ignore_errors: bool = False,
            ipython_shell: bool = False,
            out_channels: Optional[int] = None,
            sample_plotting_handler: Optional[Callable] = None,
            preview_plotting_handler: Optional[Callable] = None,
            mixed_precision: bool = False,
            tqdm_kwargs: Optional[Dict] = None
    ):
        inference_kwargs = {} if inference_kwargs is None else inference_kwargs
        if preview_batch is not None and (
                'tile_shape' not in inference_kwargs or (
                    'overlap_shape' not in inference_kwargs and 'offset' not in inference_kwargs)):
            raise ValueError(
                'If preview_batch is set, you will also need to specify '
                'tile_shape and overlap_shape or offset in inference_kwargs!'
            )
        if knossos_preview_config is not None:
            if preview_batch is not None:
                raise ValueError('If you set a preview_knossos_source, you cannot also set a preview batch.')
            preview_batch = create_preview_batch_from_knossos(knossos_preview_config)
        if enable_save_trace:
            logger.warning('enable_save_trace is deprecated. Please use the save_jit option instead.')
            assert save_jit in [None, 'trace']
            save_jit = 'trace'

        # Ensure that all nn.Modules are on the right device
        model.to(device)
        if isinstance(criterion, torch.nn.Module):
            criterion.to(device)
        if isinstance(ss_criterion, torch.nn.Module):
            ss_criterion.to(device)

        self.ignore_errors = ignore_errors
        self.ipython_shell = ipython_shell
        self.device = device
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.train_dataset = train_dataset
        self.valid_dataset = valid_dataset
        self.unlabeled_dataset = unlabeled_dataset
        self.valid_metrics = valid_metrics
        self.ss_criterion = ss_criterion
        self.preview_batch = preview_batch
        self.knossos_preview_config = knossos_preview_config
        self.preview_interval = preview_interval
        self.inference_kwargs = inference_kwargs
        self.extra_save_steps = extra_save_steps
        self.overlay_alpha = overlay_alpha
        self.save_root = os.path.expanduser(save_root)
        self.example_input = example_input
        self.save_jit = save_jit
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.sample_plotting_handler = sample_plotting_handler
        self.preview_plotting_handler = preview_plotting_handler
        self.mixed_precision = mixed_precision
        self.tqdm_kwargs = {} if tqdm_kwargs is None else tqdm_kwargs

        self._tracker = HistoryTracker()
        self._timer = Timer()
        self._first_plot = True
        self._shell_info = dedent("""
            Entering IPython training shell. To continue, hit Ctrl-D twice.
            To terminate, set self.terminate = True and then hit Ctrl-D twice.
        """).strip()

        self.inference_kwargs.setdefault('batch_size', 1)
        self.inference_kwargs.setdefault('verbose', True)
        self.inference_kwargs.setdefault('apply_softmax', True)

        if self.unlabeled_dataset is not None and self.ss_criterion is None:
            raise ValueError('If an unlabeled_dataset is supplied, you must also set ss_criterion.')

        if hparams is None:
            hparams = {}
        else:
            for k, v in hparams.items():
                if isinstance(v, (tuple, list)):
                    # Convert to str because tensorboardX doesn't support
                    # tuples and lists in add_hparams()
                    hparams[k] = str(v)
        self.hparams = hparams

        self.scaler = amp.GradScaler(enabled=self.mixed_precision)

        if exp_name is None:  # Auto-generate a name based on model name and ISO timestamp
            timestamp = datetime.datetime.now().strftime('%y-%m-%d_%H-%M-%S')
            exp_name = model.__class__.__name__ + '__' + timestamp
        self.exp_name = exp_name
        self.save_path = os.path.join(save_root, exp_name)
        if os.path.isdir(self.save_path):
            raise RuntimeError(
                f'{self.save_path} already exists.\nPlease choose a '
                'different combination of save_root and exp_name.'
            )
        os.makedirs(self.save_path)
        _change_log_file_to(f'{self.save_path}/elektronn3.log')
        logger.info(f'Writing files to save_path {self.save_path}/\n')

        self.terminate = False
        self.step = 0
        self.epoch = 0
        if schedulers is None:
            schedulers = {'lr': StepLR(optimizer, 1000, 1)}  # No-op scheduler
        self.schedulers = schedulers
        self.__lr_closetozero_alreadytriggered = False  # Used in periodic scheduler handling
        self._lr_nhood = deque(maxlen=3)  # Keeps track of the last, current and next learning rate

        self.out_channels = out_channels
        self.max_plot_id = None

        try:
            self.max_plot_id = max(self.out_channels, self.criterion.ignore_index + 1)
        except AttributeError: # no ignore_idx
            self.max_plot_id = self.out_channels
        except TypeError: # no out_channels
            self.max_plot_id = None

        if enable_videos:
            try:
                import moviepy
            except:
                logger.warning('moviepy is not installed. Disabling video logs.')
                enable_videos = False
        self.enable_videos = enable_videos
        self.tb = None  # Tensorboard handler
        if enable_tensorboard:
            if self.sample_plotting_handler is None:
                self.sample_plotting_handler = handlers._tb_log_sample_images
            if self.preview_plotting_handler is None:
                self.preview_plotting_handler = handlers._tb_log_preview

            if tensorboard_root_path is None:
                tb_path = self.save_path
            else:
                tensorboard_root_path = os.path.expanduser(tensorboard_root_path)
                tb_path = os.path.join(tensorboard_root_path, self.exp_name)
                os.makedirs(tb_path, exist_ok=True)
            self.tb = tensorboardX.SummaryWriter(logdir=tb_path, flush_secs=20)

            if self.hparams:
                self.tb.add_hparams(hparam_dict=self.hparams, metric_dict={})

        self.train_loader = DataLoader(
            self.train_dataset, batch_size=self.batch_size, shuffle=True,
            num_workers=self.num_workers, pin_memory=True,
            timeout=60 if self.num_workers > 0 else 0,
            worker_init_fn=_worker_init_fn
        )
        if valid_dataset is not None:
            self.valid_loader = DataLoader(
                self.valid_dataset, self.batch_size, shuffle=True, num_workers=self.num_workers,
                pin_memory=True, worker_init_fn=_worker_init_fn
            )
        if self.unlabeled_dataset is not None:
            self.unlabeled_loader = DataLoader(
                self.unlabeled_dataset, batch_size=self.batch_size, shuffle=True,
                num_workers=self.num_workers, pin_memory=True,
                timeout=60 if self.num_workers > 0 else 0, worker_init_fn=_worker_init_fn
            )

        self.best_val_loss = np.inf  # Best recorded validation loss
        self.best_tr_loss = np.inf

        self.valid_metrics = {} if valid_metrics is None else valid_metrics

    def run(self, max_steps: int = 1, max_runtime=3600 * 24 * 7) -> None:
        """Train the network for ``max_steps`` steps.
        After each training epoch, validation performance is measured and
        visualizations are computed and logged to tensorboard."""
        self.start_time = datetime.datetime.now()
        self.end_time = self.start_time + datetime.timedelta(seconds=max_runtime)
        self._save_model(suffix='_initial', verbose=False)
        self._lr_nhood.clear()
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])  # LR of the first training step
        while not self.terminate:
            try:
                stats, misc, tr_sample_images = self._train(max_steps, max_runtime)
                self.epoch += 1

                if self.valid_dataset is None:
                    stats['val_loss'] = nan
                    val_sample_images = None
                else:
                    valid_stats, val_sample_images = self._validate()
                    stats.update(valid_stats)

                # Log to stdout and text log file
                self._log_basic(stats, misc)
                # Render visualizations and log to tensorboard
                self._log_to_tensorboard(stats, misc, tr_sample_images, val_sample_images)
                # Legacy non-tensorboard logging to files
                self._log_to_history_tracker(stats, misc)

                # Save trained model state
                self._save_model(val_loss=stats['val_loss'], verbose=False)  # Not verbose because it can get spammy.
                # TODO: Support other metrics for determining what's the "best" model?
                if stats['val_loss'] < self.best_val_loss:
                    self.best_val_loss = stats['val_loss']
                    self._save_model(suffix='_best', val_loss=stats['val_loss'])
            except KeyboardInterrupt:
                if self.ipython_shell:
                    IPython.embed(header=self._shell_info)
                else:
                    break
                if self.terminate:
                    break
            except Exception as e:
                logger.exception('Unhandled exception during training:')
                if self.ignore_errors:
                    # Just print the traceback and try to carry on with training.
                    # This can go wrong in unexpected ways, so don't leave the training unattended.
                    pass
                elif self.ipython_shell:
                    print("\nEntering Command line such that Exception can be "
                          "further inspected by user.\n\n")
                    IPython.embed(header=self._shell_info)
                    if self.terminate:
                        break
                else:
                    raise e
        self._save_model(suffix='_final')
        if self.tb is not None:
            self.tb.close()  # Ensure that everything is flushed

    def _train_step(self, batch: Dict[str, Any]) -> Tuple[torch.Tensor, torch.Tensor]:
        """Core training step on self.device"""
        inp = batch['inp']
        target = batch.get('target')
        target_class = batch.get('class')
        # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
        dinp = inp.to(self.device, non_blocking=True)
        dtarget = target.to(self.device, non_blocking=True) if target is not None else None
        dtarget_class = target_class.to(self.device, non_blocking=True) if target_class is not None else None
        # forward pass
        with amp.autocast(enabled=self.mixed_precision):
            dout = self.model(dinp)
            if dtarget_class is not None:
                dloss = self.criterion(dout, dtarget, dtarget_class)
            else:
                dloss = self.criterion(dout, dtarget)

        unlabeled = batch.get('unlabeled')
        if unlabeled is not None:  # Add a simple consistency loss
            u_inp = unlabeled['inp']
            du_inp = u_inp.to(self.device, non_blocking=True)
            with amp.autocast(enabled=self.mixed_precision):
                du_loss = self.ss_criterion(du_inp)
                dloss += du_loss
            self.tb.add_scalar('stats/tr_uloss', float(du_loss), global_step=self.step)

        if torch.isnan(dloss):
            logger.error('NaN loss detected! Aborting training.')
            raise NaNException
        # update step
        self.scaler.scale(dloss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad(set_to_none=True)
        return dloss, dout

    def _train(self, max_steps, max_runtime):
        """Train for one epoch or until max_steps or max_runtime is reached"""
        self.model.train()

        # Scalar training stats that should be logged and written to tensorboard later
        stats: Dict[str, Union[float, List[float]]] = {stat: [] for stat in ['tr_loss']}
        # Other scalars to be logged
        misc: Dict[str, Union[float, List[float]]] = {misc: [] for misc in ['mean_target']}
        # Hold image tensors for real-time training sample visualization in tensorboard
        images: Dict[str, np.ndarray] = {}

        running_vx_size = 0  # Counts input sizes (number of pixels/voxels) of training batches
        timer = Timer()
        batch_iter = tqdm(
            self.train_loader,
            'Training',
            total=len(self.train_loader),
            dynamic_ncols=True,
            **self.tqdm_kwargs
        )
        unlabeled_iter = None if self.unlabeled_dataset is None else iter(self.unlabeled_loader)
        for i, batch in enumerate(batch_iter):
            if self.step in self.extra_save_steps:
                self._save_model(f'_step{self.step}', verbose=True)

            if unlabeled_iter is not None:
                batch['unlabeled'] = next(unlabeled_iter)
            dloss, dout = self._train_step(batch)

            with torch.no_grad():
                loss = float(dloss)
                target = batch.get('target')
                mean_target = float(target.to(torch.float32).mean()) if target is not None else 0.
                misc['mean_target'].append(mean_target)
                stats['tr_loss'].append(loss)
                batch_iter.set_description(f'Training (loss {loss:.4f})')
                self._tracker.update_timeline([self._timer.t_passed, loss, mean_target])

            # Not using .get_lr()[-1] because ReduceLROnPlateau does not implement get_lr()
            misc['learning_rate'] = self.optimizer.param_groups[0]['lr']  # LR for the this iteration
            self._scheduler_step(loss)

            running_vx_size += batch['inp'].numel()

            self._incr_step(max_runtime, max_steps)
            if i == len(self.train_loader) - 1 or self.terminate:
                # Last step in this epoch or in the whole training
                # Preserve last training batch and network output for later visualization
                images['inp'] = batch['inp'].numpy()
                images['fname'] = batch.get('fname')
                if 'target' in batch:
                    images['target'] = batch['target'].numpy()
                if 'unlabeled' in batch:
                    images['unlabeled'] = batch['unlabeled']
                images['out'] = dout.detach().cpu().numpy()
                self._put_current_attention_maps_into(images)

            if self.terminate:
                break

        stats['tr_loss_std'] = np.std(stats['tr_loss'])
        misc['tr_speed'] = len(self.train_loader) / timer.t_passed
        misc['tr_speed_vx'] = running_vx_size / timer.t_passed / 1e6  # MVx

        return stats, misc, images

    def _put_current_attention_maps_into(self, images):
        if getattr(self.model, 'attention', None):
            for i in range(len(self.model.up_convs)):
                att = self.model.up_convs[i].att[0][0].detach().cpu().numpy()
                if att.ndim == 3:
                    att = att[att.shape[0] // 2]
                images[f'att{i}'] = att

    def _incr_step(self, max_runtime, max_steps):
        """Increment the current training step counter"""
        self.step += 1
        if self.step >= max_steps:
            logger.info(f'max_steps ({max_steps}) exceeded. Terminating...')
            self.terminate = True
        if datetime.datetime.now() >= self.end_time:
            logger.info(f'max_runtime ({max_runtime} seconds) exceeded. Terminating...')
            self.terminate = True

    def _scheduler_step(self, loss):
        """Update schedules"""
        for sched in self.schedulers.values():
            # support ReduceLROnPlateau; doc. uses validation loss instead
            # http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau
            if 'metrics' in inspect.signature(sched.step).parameters:
                sched.step(metrics=loss)
            else:
                sched.step()
        # Append LR of the next iteration (after sched.step()) for local LR minima detection
        self._lr_nhood.append(self.optimizer.param_groups[0]['lr'])
        self._handle_lr()

    def _handle_lr(self) -> None:
        r"""Handle quasi-periodic learning rate schedulers that lower the
        learning rate to local minima but then ramp it up again
        (Cosine Annealing, SGDR, Cyclical LRs etc.).

        Model saving is triggered when a local minimum of learning rates is
        detected. For the motivation of this behavior, see
        https://arxiv.org/abs/1704.00109. The saved models can be used to build
        an ensemble.

        Local minima are found by checking for the simple criterion
        :math:`\lr_{t-1}` > \lr{t} < lr{t+1}`.

        If an SWA (Stochastic Weight Averaging) optimizer is detected, the SWA
        algorithm is performed (see https://arxiv.org/abs/1803.05407) and the
        resulting model is also saved, marked by the "_swa" file name suffix.

        .. note::
            The saved SWA model performs batch norm statistics correction
            only on a limited number of batches from the ``self.train_loader``
            (currently hardcoded to 10),
            so if the model uses batch normalization with running statistics and
            you suspect that this amount of batches won't be representative
            enough for your input data distribution, you might want to ensure a
            good estimate yourself by running
            :py:meth:`elektronn3.trainer.SWA.bn_update()` on the model with a
            larger number of input batches after loading the model for
            inference.
        """
        if len(self._lr_nhood) < 3:
            return  # Can't get lrs, but at this early stage it's also not relevant
        last_lr = self._lr_nhood[-3]
        curr_lr = self._lr_nhood[-2]
        next_lr = self._lr_nhood[-1]
        if last_lr > curr_lr < next_lr:
            logger.info(
                f'Local learning rate minimum {curr_lr:.2e} detected at step '
                f'{self.step}. Saving model...')
            self._save_model(suffix=f'_minlr_step{self.step}')
            # Handle Stochastic Weight Averaging optimizer if SWA is used
            if isinstance(self.optimizer, SWA):
                # TODO: Make bn_update configurable (esp. number of batches)
                self.optimizer.update_swa()  # Put current model params into SWA buffer
                self.optimizer.swap_swa_sgd()  # Perform SWA and write results into model params
                has_bn = any(isinstance(m, torch.nn.modules.batchnorm._BatchNorm) for m in self.model.modules())
                if has_bn:  # Perform batch norm correction
                    try:
                        max_bn_corr_batches = 10  # Batches to use to correct SWA batchnorm stats
                        # We're assuming here that len(self.train_loader), which is an upper bound for
                        #  len(swa_loader), is sufficient for a good stat estimation
                        swa_loader = islice(self.train_loader, max_bn_corr_batches)
                        # This may be expensive (comparable to validation computations)
                        SWA.bn_update(swa_loader, self.model, device=self.device)
                        self._save_model(suffix='_swa', verbose=False)
                    except:
                        logger.exception(
                            'SWA helper bn_update has failed. SWA model will be saved with incorrect '
                            'batchnorm statistics. Please make sure to manually correct the BN stats '
                            'before deploying the model.'
                        )
                        self._save_model(suffix='_swa_todo_batchnorm_corr', verbose=False)
                else:  # No batch norm -> save model directly
                    self._save_model(suffix='_swa', verbose=False)
                self.optimizer.swap_swa_sgd()  # Swap back model to the original state before SWA

    @torch.no_grad()
    def _validate(self) -> Tuple[Dict[str, float], Dict[str, np.ndarray]]:
        self.model.eval()  # Set dropout and batchnorm to eval mode

        val_loss = []
        outs = []
        targets = []
        stats = {name: [] for name in self.valid_metrics.keys()}
        batch_iter = tqdm(
            enumerate(self.valid_loader),
            'Validating',
            total=len(self.valid_loader),
            dynamic_ncols=True,
            **self.tqdm_kwargs
        )
        for i, batch in batch_iter:
            # Everything with a "d" prefix refers to tensors on self.device (i.e. probably on GPU)
            inp = batch['inp']
            target = batch.get('target')
            target_class = batch.get('class')
            dinp = inp.to(self.device, non_blocking=True)
            dtarget = target.to(self.device, non_blocking=True) if target is not None else None
            dtarget_class = target_class.to(self.device, non_blocking=True) if target_class is not None else None
            with amp.autocast(enabled=self.mixed_precision):
                dout = self.model(dinp)
                if dtarget is None:  # Use self-supervised unary loss function
                    val_loss.append(self.ss_criterion(dout).item())
                elif dtarget_class is not None:
                    val_loss.append(self.criterion(dout, dtarget, dtarget_class).item())
                else:
                    val_loss.append(self.criterion(dout, dtarget).item())
            out = dout.detach().cpu()
            outs.append(out)
            targets.append(target)

        images = {
            'inp': inp.numpy(),
            'out': out.numpy(),
            'target': None if target is None else target.numpy(),
            'fname': batch.get('fname'),
        }
        self._put_current_attention_maps_into(images)

        stats['val_loss'] = np.mean(val_loss)
        stats['val_loss_std'] = np.std(val_loss)

        for name, evaluator in self.valid_metrics.items():
            mvals = [evaluator(target, out) for target, out in zip(targets, outs)]
            if np.all(np.isnan(mvals)):
                stats[name] = np.nan
            else:
                stats[name] = np.nanmean(mvals)

        # # This code is currently commented out because it's quite slow. TODO: Speed up by computing softmax on GPU above
        # # Plot per-class PR curves if a supported classification scenario is detected.
        # if out.ndim == target.ndim + 1 and self.inference_kwargs.get('apply_softmax'):
        #     softmax_outs = torch.stack(outs).softmax(2)  # Apply softmax in dim=2 because of additional stack dim
        #     for c in range(out.shape[1]):
        #         self.tb.add_pr_curve(
        #             f'pr_c{c}',
        #             labels=torch.stack(targets),
        #             predictions=torch.stack([so[:, c] for so in softmax_outs]),
        #             global_step=self.step
        #         )

        self.model.train()  # Reset model to training mode

        return stats, images

# TODO: Instead of using specific keys like val_loss, enable passing info as an
#       extra dict whose contents will be added to the state_dict
    def _save_model(
            self,
            suffix: str = '',
            unwrap_parallel: bool = True,
            verbose: bool = True,
            val_loss=np.nan
    ) -> None:
        """Save/serialize trained model state to files.

        Writes the following files in the ``self.save_path``:

        - ``state_dicts.pth`` contains the a dict that holds the ``state_dict``
          of the trained model, the ``state_dict`` of the optimizer and
          some meta information (global step, epoch, best validation loss)
          The included parameters can be read and used to overwrite another
          model's ``state_dict``. The model code (architecture) itself is not
          included in this file.
        - ``model.pt`` contains a pickled version of the complete model,
          including the trained weights. You can simply
          ``model = torch.load('model.pt')`` to obtain the full model and its
          training state. This will not work if the source code relevant to de-
          serializing the model object has changed! If this is is the case,
          you will need to use the ``state_dict.pth`` to extract parameters and
          manually load them into a model.
        - ``model.pts`` contains the model in the ``torch.jit`` ScriptModule
          serialization format. If ``model`` is not already a ``ScriptModule``
          and ``self.save_jit`` is not ``None``, a ScriptModule form of the
          ``model`` will be created on demand.

        Args:
            suffix: If defined, this string will be added before the file
                extensions of the respective files mentioned above.
            unwrap_parallel: If ``True`` (default) and the model uses a parallel
                module wrapper like ``torch.nn.DataParallel``, this is
                automatically detected and the wrapped model is saved directly
                to make later deserialization easier. This can be disabled by
                setting ``unwrap_parallel=False``.
            verbose: If ``True`` (default), log infos about saved models at
                log-level "INFO" (which appears in stdout). Else, only silently
                log with log-level "DEBUG".
            val_loss: Stores the validation loss
                (default value if not supplied: NaN)
        """
        log = logger.info if verbose else logger.debug

        model = self.model

        model_trainmode = model.training

        # We do this awkard check because there are too many different
        # parallel wrappers in PyTorch and some of them have changed names
        # in different releases (DataParallel, DistributedDataParallel{,CPU}).
        is_wrapped = (
            hasattr(model, 'module') and
            'parallel' in str(type(model)).lower() and
            isinstance(model.module, torch.nn.Module)
        )
        if is_wrapped and unwrap_parallel:
            # If a parallel wrapper was used, the only thing we should save
            # is the model.module, which contains the actual model and params.
            # If we saved the wrapped module directly, deserialization would
            # get unnecessarily difficult.
            model = model.module

        state_dict_path = os.path.join(self.save_path, f'state_dict{suffix}.pth')
        model_path = os.path.join(self.save_path, f'model{suffix}.pt')

        try:
            lr_sched_state = self.schedulers['lr'].state_dict()
        except:  # No valid scheduler in use
            lr_sched_state = None

        info = {
            'global_step': self.step,
            'epoch': self.epoch,
            'best_val_loss': self.best_val_loss,
            'val_loss': val_loss,
            'inference_kwargs': self.inference_kwargs,
            'elektronn3.__version__': elektronn3.__version__,
            'env_info': collect_env.get_pretty_env_info()
        }
        # Make sure everything is a string (if inference_kwargs contains a
        #  transform object, it may not be picklable)
        info = {k: str(v) for k, v in info.items()}

        torch.save({
            'model_state_dict': model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'lr_sched_state_dict': lr_sched_state,
            'scaler_state_dict': self.scaler.state_dict(),
            'info': info
        }, state_dict_path)
        log(f'Saved state_dict as {state_dict_path}')
        pts_model_path = f'{model_path}s'
        try:
            # Try saving directly as an uncompiled nn.Module
            torch.save(model, model_path)
            log(f'Saved model as {model_path}')
            if self.save_jit == 'script':  # Compile directly for serialization
                jitmodel = torch.jit.script(model)
            elif self.save_jit == 'trace':  # Trace and serialize the model in eval mode
                if self.example_input is None:
                    raise ValueError('If save_jit="trace", example_input needs to be specified.')
                with warnings.catch_warnings():
                    # It's enough to be warned once during initial tracing
                    warnings.filterwarnings("ignore", category=torch.jit.TracerWarning)
                    jitmodel = torch.jit.trace(model.eval(), self.example_input.to(self.device))
            if self.save_jit is not None:  # Save jit model, either from script or trace
                jitmodel.save(pts_model_path)
                log(f'Saved jitted model ({self.save_jit}) as {pts_model_path}')
        except (TypeError, PickleError) as exc:
            # If model is already a ScriptModule, it can't be saved with torch.save()
            # Use ScriptModule.save() instead in this case.
            # Using the file extension '.pts' to show it's a ScriptModule.
            if isinstance(model, torch.jit.ScriptModule):
                model_path += 's'
                model.save(pts_model_path)
                log(f'Saved jitted model as {pts_model_path}')
            else:
                raise exc
        finally:
            # Reset training state to the one it had before this function call,
            # because it could have changed with the model.eval() call above.
            model.training = model_trainmode
        if os.path.isfile(pts_model_path):
            with zipfile.ZipFile(pts_model_path, 'a', compression=zipfile.ZIP_DEFLATED) as zfile:
                infostr = pprint.pformat(info, indent=2, width=120)
                zfile.writestr('info.txt', infostr)

    def _log_basic(self, stats, misc):
        """Log to stdout and text log file"""
        tr_loss = np.mean(stats['tr_loss'])
        val_loss = np.mean(stats['val_loss'])
        lr = misc['learning_rate']
        tr_speed = misc['tr_speed']
        tr_speed_vx = misc['tr_speed_vx']
        t = pretty_string_time(self._timer.t_passed)
        text = f'step={self.step:06d}, tr_loss={tr_loss:.3f}, val_loss={val_loss:.3f}, '
        text += f'lr={lr:.2e}, {tr_speed:.2f} it/s, {tr_speed_vx:.2f} MVx/s, {t}'
        logger.info(text)

    def _log_to_tensorboard(
            self,
            stats: Dict,
            misc: Dict,
            tr_images: Dict,
            val_images: Optional[Dict] = None,
            file_stats: Optional[Dict] = None,
    ) -> None:
        """Create visualizations, make preview predictions, log and plot to tensorboard"""
        if self.tb:
            try:
                self._tb_log_scalars(stats, 'stats')
                self._tb_log_scalars(misc, 'misc')
                if self.preview_batch is not None:
                    if self.epoch % self.preview_interval == 0 or self.epoch == 1:
                        # TODO: Also save preview inference results in a (3D) HDF5 file
                        self.preview_plotting_handler(self)
                self.sample_plotting_handler(self, tr_images, group='tr_samples')
                if val_images is not None:
                    self.sample_plotting_handler(self, val_images, group='val_samples')
                if file_stats is not None:
                    self._tb_log_scalars(file_stats, 'file_stats')
                self._tb_log_histograms()
            except Exception:
                logger.exception('Error occured while logging to tensorboard:')

    def _log_to_history_tracker(self, stats: Dict, misc: Dict) -> None:
        """Update history tracker and plot stats (kind of made obsolete by tensorboard)"""
        # TODO: Decide what to do with this, now that most things are already in tensorboard.
        if self._tracker.history.length > 0:
            tr_loss_gain = self._tracker.history[-1][2] - np.mean(stats['tr_loss'])
        else:
            tr_loss_gain = 0
        if not stats.get('tr_accuracy'):
            tr_accuracy = nan
        else:
            tr_accuracy = np.nanmean(stats['tr_accuracy'])
        val_accuracy = stats.get('val_accuracy', nan)
        self._tracker.update_history([
            self.step, self._timer.t_passed, np.mean(stats['tr_loss']), np.mean(stats['val_loss']),
            tr_loss_gain, tr_accuracy, val_accuracy, misc['learning_rate'], 0, 0
        ])
        # Plot tracker stats to pngs in save_path
        self._tracker.plot(self.save_path)

    def _tb_log_scalars(
            self,
            scalars: Dict[str, float],
            tag: str = 'default'
    ) -> None:
        for key, value in scalars.items():
            if isinstance(value, (list, tuple, np.ndarray)):
                for i in range(len(value)):
                    if not np.isnan(value[i]):
                        self.tb.add_scalar(f'{tag}/{key}', value[i], self.step - len(value) + i)
            elif not np.isnan(value):
                self.tb.add_scalar(f'{tag}/{key}', value, self.step)

    def _tb_log_histograms(self) -> None:
        """Log histograms of model parameters and their current gradients.

        Make sure to run this between ``backward()`` and ``zero_grad()``,
        because otherwise gradient histograms will only consist of zeros.
        """
        for name, param in self.model.named_parameters():
            self.tb.add_histogram(f'param/{name}', param, self.step)
            grad = param.grad if param.grad is not None else torch.tensor(0)
            self.tb.add_histogram(f'grad/{name}', grad, self.step)

    def _preview_inference(
            self,
            inp: np.ndarray,
            inference_kwargs: Dict[str, Any],
    ) -> torch.Tensor:
        if self.out_channels is None:
            raise RuntimeError('Can\'t do preview prediction if Trainer.out_channels is not set.')
        out_shape = (self.out_channels, *inp.shape[2:])
        predictor = Predictor(
            model=self.model,
            device=self.device,
            out_shape=out_shape,
            **inference_kwargs,
        )
        out = predictor.predict(inp)
        return out


class Backup:
    """ Backup class for archiving training script, src folder and environment info.
    Should be used for any future archiving needs.

    Args:
        script_path: The path to the training script. Eg. train_unet_neurodata.py
        save_path: The path where the information is archived.
        extra_content: Dictionary of {filename: content} entries, where content
            is a string that should be written to a file with the specified name.

    """
    def __init__(self, script_path, save_path, extra_content=None):
        self.script_path = script_path
        self.save_path = save_path
        self.extra_content = extra_content

    def archive_backup(self):
        """Archiving the source folder, the training script and environment info.

        The training script is saved with the prefix "0-" to distinguish from regular scripts.
        Environment information equivalent to the output of ``python -m torch.utils.collect_env``
        is saved in a file named "env_info.txt".
        """

        # Archiving the Training script
        shutil.copyfile(self.script_path, self.save_path + '/0-' + os.path.basename(self.script_path))
        os.chmod(self.save_path + '/0-' + os.path.basename(self.script_path), 0o755)
        # Archiving the src folder
        pkg_path = os.path.dirname(arch_src)
        backup_path = os.path.join(self.save_path, 'src_backup')
        shutil.make_archive(backup_path, 'gztar', pkg_path)

        # Archiving the Environment Info
        env_info = collect_env.get_pretty_env_info()
        with open(self.save_path + '/env_info.txt', 'w') as f:
            f.write(env_info)
        if self.extra_content is not None:
            for fname, content in self.extra_content.items():
                with open(f'{self.save_path}/{fname}', 'w') as f:
                    f.write(content)


def findcudatensors() -> Tuple[int, List[torch.Tensor]]:
    """Find currently living tensors that are allocated on cuda device memory.
    This can be used for debugging memory leaks:
    If ``findcudatensors()[0]`` grows unexpectedly between GPU computations,
    you can look at the returned ``tensors`` list to find out what tensors
    are currently allocated, for example
    ``print([x.shape for x in findcudatensors()[1])``.

    Returns a tuple of

    - total memory usage of found tensors in MiB
    - a list of all of those tensors, ordered by size."""
    tensors = []
    for obj in gc.get_objects():
        try:
            if torch.is_tensor(obj) \
                    and obj.device.type == 'cuda' \
                    and not isinstance(obj, torch.nn.Parameter):  # Exclude model params
                tensors.append(obj)
        except:
            pass
    tensors.sort(key=lambda x: x.numel())
    total_mib = sum(x.numel() * 32 for x in tensors) / 1024**2  # Assuming float32
    return total_mib, tensors



# This implementation of V-Net (https://arxiv.org/abs/1606.04797)
# is taken from https://github.com/mattmacy/vnet.pytorch by Matthew Macy
# It is published under the BSD 3-Clause License. (Thanks, Matthew.)

import torch
import torch.nn as nn
import torch.nn.functional as F


def passthrough(x, **kwargs):
    return x


def ELUCons(relu, nchan):
    if relu:
        return nn.ReLU(inplace=True)
    else:
        return nn.PReLU(nchan)

# normalization between sub-volumes is necessary
# for good performance
class ContBatchNorm3d(nn.modules.batchnorm._BatchNorm):
    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'
                             .format(input.dim()))

    def forward(self, input):
        self._check_input_dim(input)
        return F.batch_norm(
            input, self.running_mean, self.running_var, self.weight, self.bias,
            True, self.momentum, self.eps)


class LUConv(nn.Module):
    def __init__(self, nchan, relu):
        super(LUConv, self).__init__()
        self.relu1 = ELUCons(relu, nchan)
        self.conv1 = nn.Conv3d(nchan, nchan, kernel_size=5, padding=2)
        self.bn1 = ContBatchNorm3d(nchan)

    def forward(self, x):
        x = self.relu1(self.bn1(self.conv1(x)))
        return x


def _make_nConv(nchan, depth, relu):
    layers = []
    for _ in range(depth):
        layers.append(LUConv(nchan, relu))
    return nn.Sequential(*layers)


class InputTransition(nn.Module):
    def __init__(self, outChans, relu):
        super(InputTransition, self).__init__()
        self.conv1 = nn.Conv3d(1, outChans, kernel_size=5, padding=2)
        self.bn1 = ContBatchNorm3d(outChans)
        self.relu1 = ELUCons(relu, outChans)

    def forward(self, x):
        out = self.bn1(self.conv1(x))
        out = self.relu1(out)
        return out


class DownTransition(nn.Module):
    def __init__(self, inChans, nConvs, relu, dropout=False):
        super(DownTransition, self).__init__()
        outChans = 2*inChans
        self.down_conv = nn.Conv3d(inChans, outChans, kernel_size=2, stride=2)
        self.bn1 = ContBatchNorm3d(outChans)
        self.do1 = passthrough
        self.relu1 = ELUCons(relu, outChans)
        self.relu2 = ELUCons(relu, outChans)
        if dropout:
            self.do1 = nn.Dropout3d()
        self.ops = _make_nConv(outChans, nConvs, relu)

    def forward(self, x):
        down = self.relu1(self.bn1(self.down_conv(x)))
        x = self.do1(down)
        x = self.ops(x)
        x = self.relu2(torch.add(x, down))
        return x


class UpTransition(nn.Module):
    def __init__(self, inChans, outChans, nConvs, relu, dropout=False):
        super(UpTransition, self).__init__()
        self.up_conv = nn.ConvTranspose3d(inChans, outChans // 2, kernel_size=2, stride=2)
        self.bn1 = ContBatchNorm3d(outChans // 2)
        self.do1 = passthrough
        self.do2 = nn.Dropout3d()
        self.relu1 = ELUCons(relu, outChans // 2)
        self.relu2 = ELUCons(relu, outChans)
        if dropout:
            self.do1 = nn.Dropout3d()
        self.ops = _make_nConv(outChans, nConvs, relu)

    def forward(self, x, skipx):
        x = self.do1(x)
        skipxdo = self.do2(skipx)
        x = self.relu1(self.bn1(self.up_conv(x)))
        xcat = torch.cat((x, skipxdo), 1)
        x = self.ops(xcat)
        x = self.relu2(torch.add(x, xcat))
        return x


class OutputTransition(nn.Module):
    def __init__(self, inChans, relu, nll):
        super(OutputTransition, self).__init__()
        self.conv1 = nn.Conv3d(inChans, 2, kernel_size=1, )
        self.bn1 = ContBatchNorm3d(2)
        self.relu1 = ELUCons(relu, 2)

    def forward(self, x):
        # convolve 32 down to 2 channels
        x = self.relu1(self.bn1(self.conv1(x)))
        return x


class VNet(nn.Module):
    # the number of convolutions in each layer corresponds
    # to what is in the actual prototxt, not the intent
    def __init__(self, relu=True, nll=True, fac=4):
        super(VNet, self).__init__()
        self.in_tr = InputTransition(16 // fac, relu)
        self.down_tr32 = DownTransition(16 // fac, 1, relu)
        self.down_tr64 = DownTransition(32 // fac, 2, relu)
        self.down_tr128 = DownTransition(64 // fac, 3, relu, dropout=True)
        self.down_tr256 = DownTransition(128 // fac, 2, relu, dropout=True)
        self.up_tr256 = UpTransition(256 // fac, 256 // fac, 2, relu, dropout=True)
        self.up_tr128 = UpTransition(256 // fac, 128 // fac, 2, relu, dropout=True)
        self.up_tr64 = UpTransition(128 // fac, 64 // fac, 1, relu)
        self.up_tr32 = UpTransition(64 // fac, 32 // fac, 1, relu)
        self.out_tr = OutputTransition(32 // fac, relu, nll)

    # The network topology as described in the diagram
    # in the VNet paper
    # def __init__(self):
    #     super(VNet, self).__init__()
    #     self.in_tr =  InputTransition(16)
    #     # the number of convolutions in each layer corresponds
    #     # to what is in the actual prototxt, not the intent
    #     self.down_tr32 = DownTransition(16, 2)
    #     self.down_tr64 = DownTransition(32, 3)
    #     self.down_tr128 = DownTransition(64, 3)
    #     self.down_tr256 = DownTransition(128, 3)
    #     self.up_tr256 = UpTransition(256, 3)
    #     self.up_tr128 = UpTransition(128, 3)
    #     self.up_tr64 = UpTransition(64, 2)
    #     self.up_tr32 = UpTransition(32, 1)
    #     self.out_tr = OutputTransition(16)
    def forward(self, x):
        out16 = self.in_tr(x)
        out32 = self.down_tr32(out16)
        out64 = self.down_tr64(out32)
        out128 = self.down_tr128(out64)
        out256 = self.down_tr256(out128)
        out = self.up_tr256(out256, out128)
        del out256
        del out128
        out = self.up_tr128(out, out64)
        del out64
        out = self.up_tr64(out, out32)
        del out32
        out = self.up_tr32(out, out16)
        del out16
        out = self.out_tr(out)
        return out



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Ravil Dorozhinskii

import numpy as np


class Region:
    """ Is a container that holds spatial coordinates
    of the region
    """
    def __init__(self,
                 coords_lo: list,
                 coords_hi: list,
                 size: list = None):
        """
        Parameters
        ----------
        coords_lo - the lowest region coordinates along each axis
        coords_hi - the highest region coordinates along each axis
        size - spatial sizes of the region along each axis
        """

        self.coords_lo = coords_lo
        self.coords_hi = coords_hi

        if size:
            self.size = size
        else:
            self.size = [high - low for high, low in zip(coords_hi, coords_lo)]


class RegionGenerator:
    """ A class instance generates regions with arbitrary
    spatial size and location within the specified coordinate bounds.
    The coordinate bounds are usually the spatial size of
    the input sample.
    """
    def __init__(self,
                 coord_bounds: list,
                 lower_lim_region_size: list,
                 upper_lim_region_size: list):
        """
        Parameters
        ----------
        coord_bounds - coordinate bounds of a sample
            with the format: [depth, width, height]
        lower_lim_region_size - region minimal size along each axis
            with the format: [min_depth, min_width, min_height]
        upper_lim_region_size - region maximal size along each axis
            with the format: [max_depth, max_width, max_height]
        """

        self.sample_size = coord_bounds
        self.coords_lo_lim = lower_lim_region_size
        self.coords_hi_lim = upper_lim_region_size
        self.dim = len(self.sample_size)

    def create_region(self) -> Region:
        """ Generates a region with arbitrary spatial size
        and location according to the parameters passed by the user
        to the constructor
        Returns
        -------
        instance of the Region class
        """

        size = [np.random.randint(low=self.coords_lo_lim[i],
                                  high=self.coords_hi_lim[i])
                for i in range(self.dim)]

        coords_lo = [np.random.randint(low=0,
                                       high=self.sample_size[i] - size[i])
                     for i in range(self.dim)]

        coords_hi = [coords_lo[i] + size[i] for i in range(self.dim)]

        return Region(coords_lo, coords_hi, size)



# This file is copied from https://github.com/mdraw/contrib/blob/bdf4da5/torchcontrib/optim/swa.py
# (C) torchcontrib authors
# Vendored here because torchcontrib was not conda-installable as of the time of including this.

from collections import defaultdict
from itertools import chain
from torch.optim import Optimizer
import torch
import warnings


class SWA(Optimizer):
    def __init__(self, optimizer, swa_start=None, swa_freq=None, swa_lr=None):
        r"""Implements Stochastic Weight Averaging (SWA).

        Stochastic Weight Averaging was proposed in `Averaging Weights Leads to
        Wider Optima and Better Generalization`_ by Pavel Izmailov, Dmitrii
        Podoprikhin, Timur Garipov, Dmitry Vetrov and Andrew Gordon Wilson
        (UAI 2018).

        SWA is implemented as a wrapper class taking optimizer instance as input
        and applying SWA on top of that optimizer.

        SWA can be used in two modes: automatic and manual. In the automatic
        mode SWA running averages are automatically updated every
        :attr:`swa_freq` steps after :attr:`swa_start` steps of optimization. If
        :attr:`swa_lr` is provided, the learning rate of the optimizer is reset
        to :attr:`swa_lr` at every step starting from :attr:`swa_start`. To use
        SWA in automatic mode provide values for both :attr:`swa_start` and
        :attr:`swa_freq` arguments.

        Alternatively, in the manual mode, use :meth:`update_swa` or
        :meth:`update_swa_group` methods to update the SWA running averages.

        In the end of training use `swap_swa_sgd` method to set the optimized
        variables to the computed averages.

        Args:
            optimizer (torch.optim.Optimizer): optimizer to use with SWA
            swa_start (int): number of steps before starting to apply SWA in
                automatic mode; if None, manual mode is selected (default: None)
            swa_freq (int): number of steps between subsequent updates of
                SWA running averages in automatic mode; if None, manual mode is
                selected (default: None)
            swa_lr (float): learning rate to use starting from step swa_start
                in automatic mode; if None, learning rate is not changed
                (default: None)

        Examples:
            >>> # automatic mode
            >>> base_opt = torch.optim.SGD(model.parameters(), lr=0.1)
            >>> opt = torchcontrib.optim.SWA(
            >>>                 base_opt, swa_start=10, swa_freq=5, swa_lr=0.05)
            >>> for _ in range(100):
            >>>     opt.zero_grad()
            >>>     loss_fn(model(input), target).backward()
            >>>     opt.step()
            >>> opt.swap_swa_sgd()
            >>> # manual mode
            >>> opt = torchcontrib.optim.SWA(base_opt)
            >>> for i in range(100):
            >>>     opt.zero_grad()
            >>>     loss_fn(model(input), target).backward()
            >>>     opt.step()
            >>>     if i > 10 and i % 5 == 0:
            >>>         opt.update_swa()
            >>> opt.swap_swa_sgd()

        .. note::
            SWA does not support parameter-specific values of :attr:`swa_start`,
            :attr:`swa_freq` or :attr:`swa_lr`. In automatic mode SWA uses the
            same :attr:`swa_start`, :attr:`swa_freq` and :attr:`swa_lr` for all
            parameter groups. If needed, use manual mode with
            :meth:`update_swa_group` to use different update schedules for
            different parameter groups.

        .. note::
            Call :meth:`swap_swa_sgd` in the end of training to use the computed
            running averages.

        .. note::
            If you are using SWA to optimize the parameters of a Neural Network
            containing Batch Normalization layers, you need to update the
            :attr:`running_mean` and :attr:`running_var` statistics of the
            Batch Normalization module. You can do so by using
            `torchcontrib.optim.swa.bn_update` utility.

        .. _Averaging Weights Leads to Wider Optima and Better Generalization:
            https://arxiv.org/abs/1803.05407
        .. _Improving Consistency-Based Semi-Supervised Learning with Weight
            Averaging:
            https://arxiv.org/abs/1806.05594
        """
        self._auto_mode, (self.swa_start, self.swa_freq) = \
            self._check_params(self, swa_start, swa_freq)
        self.swa_lr = swa_lr

        if self._auto_mode:
            if swa_start < 0:
                raise ValueError("Invalid swa_start: {}".format(swa_start))
            if swa_freq < 1:
                raise ValueError("Invalid swa_freq: {}".format(swa_freq))
        else:
            if self.swa_lr is not None:
                warnings.warn(
                    "Some of swa_start, swa_freq is None, ignoring swa_lr")
            # If not in auto mode make all swa parameters None
            self.swa_lr = None
            self.swa_start = None
            self.swa_freq = None

        if self.swa_lr is not None and self.swa_lr < 0:
            raise ValueError("Invalid SWA learning rate: {}".format(swa_lr))

        self.optimizer = optimizer

        self.defaults = self.optimizer.defaults
        self.param_groups = self.optimizer.param_groups
        self.state = defaultdict(dict)
        self.opt_state = self.optimizer.state
        for group in self.param_groups:
            group['n_avg'] = 0
            group['step_counter'] = 0

    @staticmethod
    def _check_params(self, swa_start, swa_freq):
        params = [swa_start, swa_freq]
        params_none = [param is None for param in params]
        if not all(params_none) and any(params_none):
            warnings.warn(
                "Some of swa_start, swa_freq is None, ignoring other")
        for i, param in enumerate(params):
            if param is not None and not isinstance(param, int):
                params[i] = int(param)
                warnings.warn("Casting swa_start, swa_freq to int")
        return not any(params_none), params

    def _reset_lr_to_swa(self):
        if self.swa_lr is None:
            return
        for param_group in self.param_groups:
            if param_group['step_counter'] >= self.swa_start:
                param_group['lr'] = self.swa_lr

    def update_swa_group(self, group):
        r"""Updates the SWA running averages for the given parameter group.

        Arguments:
            param_group (dict): Specifies for what parameter group SWA running
                averages should be updated

        Examples:
            >>> # automatic mode
            >>> base_opt = torch.optim.SGD([{'params': [x]},
            >>>             {'params': [y], 'lr': 1e-3}], lr=1e-2, momentum=0.9)
            >>> opt = torchcontrib.optim.SWA(base_opt)
            >>> for i in range(100):
            >>>     opt.zero_grad()
            >>>     loss_fn(model(input), target).backward()
            >>>     opt.step()
            >>>     if i > 10 and i % 5 == 0:
            >>>         # Update SWA for the second parameter group
            >>>         opt.update_swa_group(opt.param_groups[1])
            >>> opt.swap_swa_sgd()
        """
        for p in group['params']:
            param_state = self.state[p]
            if 'swa_buffer' not in param_state:
                param_state['swa_buffer'] = torch.zeros_like(p.data)
            buf = param_state['swa_buffer']
            virtual_decay = 1 / float(group["n_avg"] + 1)
            diff = (p.data - buf) * virtual_decay
            buf.add_(diff)
        group["n_avg"] += 1

    def update_swa(self):
        r"""Updates the SWA running averages of all optimized parameters.
        """
        for group in self.param_groups:
            self.update_swa_group(group)

    def swap_swa_sgd(self):
        r"""Swaps the values of the optimized variables and swa buffers.

        It's meant to be called in the end of training to use the collected
        swa running averages. It can also be used to evaluate the running
        averages during training; to continue training `swap_swa_sgd`
        should be called again.
        """
        for group in self.param_groups:
            for p in group['params']:
                param_state = self.state[p]
                if 'swa_buffer' not in param_state:
                    # If swa wasn't applied we don't swap params
                    warnings.warn(
                        "SWA wasn't applied to param {}; skipping it".format(p))
                    continue
                buf = param_state['swa_buffer']
                tmp = torch.empty_like(p.data)
                tmp.copy_(p.data)
                p.data.copy_(buf)
                buf.copy_(tmp)

    def step(self, closure=None):
        r"""Performs a single optimization step.

        In automatic mode also updates SWA running averages.
        """
        self._reset_lr_to_swa()
        loss = self.optimizer.step(closure)
        for group in self.param_groups:
            group["step_counter"] += 1
            steps = group["step_counter"]
            if self._auto_mode:
                if steps > self.swa_start and steps % self.swa_freq == 0:
                    self.update_swa_group(group)
        return loss

    def state_dict(self):
        r"""Returns the state of SWA as a :class:`dict`.

        It contains three entries:
            * opt_state - a dict holding current optimization state of the base
                optimizer. Its content differs between optimizer classes.
            * swa_state - a dict containing current state of SWA. For each
                optimized variable it contains swa_buffer keeping the running
                average of the variable
            * param_groups - a dict containing all parameter groups
        """
        opt_state_dict = self.optimizer.state_dict()
        swa_state = {(id(k) if isinstance(k, torch.Tensor) else k): v
                     for k, v in self.state.items()}
        opt_state = opt_state_dict["state"]
        param_groups = opt_state_dict["param_groups"]
        return {"opt_state": opt_state, "swa_state": swa_state,
                "param_groups": param_groups}

    def load_state_dict(self, state_dict):
        r"""Loads the optimizer state.

        Args:
            state_dict (dict): SWA optimizer state. Should be an object returned
                from a call to `state_dict`.
        """
        swa_state_dict = {"state": state_dict["swa_state"],
                          "param_groups": state_dict["param_groups"]}
        opt_state_dict = {"state": state_dict["opt_state"],
                          "param_groups": state_dict["param_groups"]}
        super(SWA, self).load_state_dict(swa_state_dict)
        self.optimizer.load_state_dict(opt_state_dict)
        self.opt_state = self.optimizer.state

    def add_param_group(self, param_group):
        r"""Add a param group to the :class:`Optimizer` s `param_groups`.

        This can be useful when fine tuning a pre-trained network as frozen
        layers can be made trainable and added to the :class:`Optimizer` as
        training progresses.

        Args:
            param_group (dict): Specifies what Tensors should be optimized along
            with group specific optimization options.
        """
        param_group['n_avg'] = 0
        param_group['step_counter'] = 0
        self.optimizer.add_param_group(param_group)

    @staticmethod
    def bn_update(loader, model, device=None):
        r"""Updates BatchNorm running_mean, running_var buffers in the model.

        It performs one pass over data in `loader` to estimate the activation
        statistics for BatchNorm layers in the model.

        Args:
            loader (torch.utils.data.DataLoader): dataset loader to compute the
                activation statistics on. Each data batch should be either a
                tensor, or a list/tuple whose first element is a tensor
                containing data.

            model (torch.nn.Module): model for which we seek to update BatchNorm
                statistics.

            device (torch.device, optional): If set, data will be trasferred to
                :attr:`device` before being passed into :attr:`model`.
        """
        if not _check_bn(model):
            return
        was_training = model.training
        model.train()
        momenta = {}
        model.apply(_reset_bn)
        model.apply(lambda module: _get_momenta(module, momenta))
        n = 0
        for input in loader:
            if isinstance(input, (list, tuple)):
                input = input[0]
            elif isinstance(input, dict):
                input = input['inp']
            b = input.size(0)

            momentum = b / float(n + b)
            for module in momenta.keys():
                module.momentum = momentum

            if device is not None:
                input = input.to(device)

            model(input)
            n += b

        model.apply(lambda module: _set_momenta(module, momenta))
        model.train(was_training)


# BatchNorm utils
def _check_bn_apply(module, flag):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        flag[0] = True


def _check_bn(model):
    flag = [False]
    model.apply(lambda module: _check_bn_apply(module, flag))
    return flag[0]


def _reset_bn(module):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        module.running_mean = torch.zeros_like(module.running_mean)
        module.running_var = torch.ones_like(module.running_var)


def _get_momenta(module, momenta):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        momenta[module] = module.momentum


def _set_momenta(module, momenta):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        module.momentum = momenta[module]



import os
import logging
import datetime
from pickle import PicklingError
from typing import Optional, Dict, Tuple, Union, List, Any
from tqdm import tqdm
import numpy as np
import zipfile
import pprint
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("agg")

import torch
import torch.nn
import torch.utils.data
from torch.utils.tensorboard import SummaryWriter
from torch.utils import collect_env
import torch_geometric
from sklearn.manifold import TSNE
# import tensorboardX

import elektronn3
from elektronn3.training.trainer import _change_log_file_to, NaNException
from elektronn3.training.train_utils import HistoryTracker, Timer, pretty_string_time


logger = logging.getLogger('elektronn3log')

def visualize_embeddings(x, color, epoch):
    mapping = {0: "MSN", 1: "LMAN", 2: "HVC"}

    fig, ax = plt.subplots(figsize=(10, 10))
    scatter = ax.scatter(x[:, 0], x[:, 1], c=color, s=15)
    handles, labels = scatter.legend_elements()
    labels = [mapping[int(l.split("{")[1][0])] for l in labels]
    ax.set_xticks([])
    ax.set_yticks([])
    ax.legend(handles, labels, loc="best", title="Cell types")
    ax.set_title(f"Node embeddings (t-SNE), Epoch {epoch}")

    # convert figure to numpy array
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return data.reshape(3, data.shape[0], data.shape[1])  # (3, height, width)


class GNNTrainer:

    tb: SummaryWriter
    terminate: bool
    step: int
    epoch: int
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader
    exp_name: str
    save_path: str  # Full path to where training files are stored
    out_channels: Optional[int]  # Number of channels of the network outputs

    def __init__(
        self, 
        model: torch.nn.Module,
        criterion: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        save_root: str,
        data: torch_geometric.data.Data,
        valid_metrics: Optional[Dict] = None,
        exp_name: Optional[str] = None,
        batch_size: int = 1,
        early_stopping: Optional[Any] = None,
        num_workers: int = 0,
        enable_tensorboard: bool = True,
        tensorboard_root_path: Optional[str] = None,
        out_channels: Optional[int] = None,
        tqdm_kwargs: Optional[Dict] = None,
    ):
        model.to(device)
        if isinstance(criterion, torch.nn.Module):
            criterion.to(device)

        self.device = device
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.data = data
        self.valid_metrics = valid_metrics
        self.save_root = os.path.expanduser(save_root)
        self.batch_size = batch_size
        self.early_stopping = early_stopping
        self.num_workers = num_workers
        self.tqdm_kwargs = {} if tqdm_kwargs is None else tqdm_kwargs
        
        lr = self.optimizer.state_dict()["param_groups"][0]["lr"]

        logger.info(f"Training configuration: lr={lr}, criterion={self.criterion.__class__.__name__}, optimizer={self.optimizer.__class__.__name__}, model={self.model.__class__.__name__}, early_stopping={True if self.early_stopping else False}")

        self._tracker = HistoryTracker()
        self._timer = Timer()

        if exp_name is None:
            timestamp = datetime.datetime.now().strftime('%y-%m-%d_%H-%M-%S')
            exp_name = model.__class__.__name__ + '__' + timestamp

        self.exp_name = exp_name
        self.save_path = os.path.join(save_root, self.exp_name)
        if os.path.isdir(self.save_path):
            raise RuntimeError(f"Save path {self.save_path} already exists")

        os.makedirs(self.save_path)
        _change_log_file_to(f'{self.save_path}/elektronn3.log')
        logger.info(f"Writing files to save_path {self.save_path}/\n")

        # Terminate training by setting to true
        self.terminate = False
        self.epoch = 0
        self.patience = self.early_stopping.patience if self.early_stopping else 0

        self.out_channels = out_channels
        self.tb = None
        if enable_tensorboard:
            if tensorboard_root_path is None:
                tb_path = self.save_path
            else:
                tensorboard_root_path = os.path.expanduser(tensorboard_root_path)
                tb_path = os.path.join(tensorboard_root_path, self.exp_name)
                os.makedirs(tb_path, exist_ok=True)
            self.tb = SummaryWriter(tb_path, flush_secs=20)

        self.num_nodes = self.data.x.shape[0]
        # Train and val mask for semi-supervised training
        
        logger.info(f"Training nodes: {self.data.train_mask.sum()}")
        logger.info(f"Validation nodes: {self.data.val_mask.sum()}")
        logger.info(f"Training node label rate: {(self.data.train_mask.sum() / self.num_nodes) * 100:.2f}%\n")

        assert self.data.has_isolated_nodes() == False, "Dataset has isolated nodes"
        assert self.data.is_undirected() == True, "Dataset is not undirected"
        assert self.data.has_self_loops() == False, "Dataset has self-loops"

        self.best_val_loss = np.inf
        self.best_tr_loss = np.inf

    def run(self, max_epochs: int=1, max_runtime=3600*24*7) -> None:
        self.start_time = datetime.datetime.now()
        self.end_time = self.start_time + datetime.timedelta(seconds=max_runtime)
        # Save initial model
        self._save_model(suffix="_initial", verbose=False)

        # logger.info("Saving initial node embeddings\n")
        # with torch.no_grad():
        #     out = self.model(self.data)

        # embeddings = TSNE(perplexity=20, n_iter=750).fit_transform(out.cpu().detach().numpy())
        # img_data = visualize_embeddings(embeddings, self.data.y.cpu().numpy(), self.epoch)
        # self.tb.add_image("embeddings_initial", img_data, self.epoch)

        stats: Dict[str, Union[float, List[float]]] = {stat: [] for stat in ['tr_loss', 'tr_acc']}
        misc: Dict[str, Union[float, List[float]]] = {}
        
        while not self.terminate:
            try:
                timer = Timer()
                correct = 0
                
                # Training loss
                dloss, dout = self._train()
                loss = float(dloss.item())

                # Training accuracy
                pred = torch.argmax(dout, dim=1)
                correct = torch.sum(pred[self.data.train_mask] == self.data.y[self.data.train_mask]).item()

                stats['tr_loss'].append(loss)
                stats["tr_loss_std"] = np.std(stats["tr_loss"])
                stats['tr_acc'].append(correct / self.data.train_mask.sum())
                misc["tr_speed"] = self.num_nodes / timer.t_passed

                self._tracker.update_timeline([self._timer.t_passed, loss])

                valid_stats = self._validate()
                stats.update(valid_stats)

                if self.early_stopping is not None:
                    self.early_stopping(stats["val_loss"][-1])
                    if self.early_stopping.stop:
                        logger.info(f"Validation loss did not improve for {self.patience} epochs. Terminating...")
                        self.terminate = True
    
                # Log on the console
                self._log_basic(stats, misc)

                # Log on tensorboard
                self.tb.add_scalar('stats/tr_loss', stats["tr_loss"][-1], self.epoch)
                self.tb.add_scalar('stats/tr_acc', stats['tr_acc'][-1], self.epoch)
                self.tb.add_scalar('stats/val_loss', stats['val_loss'][-1], self.epoch)
                self.tb.add_scalar('stats/val_acc', stats['val_acc'][-1], self.epoch)
                self._tb_log_histograms()

                # Save current model
                self._save_model(val_loss=stats['val_loss'][-1], verbose=False)
                # Save best model
                if stats["val_loss"][-1] < self.best_val_loss:
                    self.best_val_loss = stats["val_loss"][-1]
                    self._save_model(suffix="_best", val_loss=stats["val_loss"][-1])
                    print()

                # Increment epochs
                self._incr_epoch(max_epochs, max_runtime)

            except KeyboardInterrupt: 
                break

            except Exception as e:
                logger.exception("Unhandled exception during training:")
                raise e

        # Save final model
        self._save_model(suffix="_final")
        # logger.info("Saving final node embeddings\n")
        # with torch.no_grad():
        #     out = self.model(self.data)

        # embeddings = TSNE(perplexity=20, n_iter=750).fit_transform(out.cpu().detach().numpy())
        # img_data = visualize_embeddings(embeddings, self.data.y.cpu().numpy(), self.epoch)
        # self.tb.add_image("embeddings_final", img_data, self.epoch)

        if self.tb is not None:
            self.tb.close()  # Ensure that everything is flushed

    def _train(self) -> None:
        self.model.train()
        self.optimizer.zero_grad()
        
        dout = self.model(self.data)
        dloss = self.criterion(dout[self.data.train_mask], self.data.y[self.data.train_mask].long())
        
        if torch.isnan(dloss):
            logger.error(f"NaN loss encountered at step {self.epoch}. Aborting...")
            raise NaNException()
        
        dloss.backward()
        self.optimizer.step()
        
        return dloss, dout

    @torch.no_grad()
    def _validate(self) -> Tuple[Dict[str, float], Dict[str, np.ndarray]]:
        self.model.eval()
        stats = {stat: [] for stat in ['val_loss', 'val_acc']}

        # Validation loss
        dout = self.model(self.data)
        dloss = self.criterion(dout[self.data.val_mask], self.data.y[self.data.val_mask].long())

        # Validation accuracy
        pred = torch.argmax(dout, dim=1)
        correct = torch.sum(pred[self.data.val_mask] == self.data.y[self.data.val_mask]).item()
        
        stats["val_loss"].append(dloss.item())
        stats["val_loss_std"] = np.std(stats["val_loss"])
        stats["val_acc"].append(correct / self.data.val_mask.sum())

        return stats

    def _incr_epoch(self, max_epochs, max_runtime):
        self.epoch += 1
        
        if self.epoch >= max_epochs:
            logger.info(f"max_steps ({max_epochs}) exceeded. Terminating...")
            self.terminate = True

        if datetime.datetime.now() >= self.end_time:
            logger.info(f"max_runtime ({max_runtime} seconds) exceeded. Terminating...")
            self.terminate = True

    def _save_model(
        self, 
        suffix: str="",
        verbose: bool = True,
        val_loss=np.nan
    ) -> None:
        log = logger.info if verbose else logger.debug

        model = self.model
        model_trainmode = model.training

        state_dict_path = os.path.join(self.save_path, f"state_dict{suffix}.pth")
        model_path = os.path.join(self.save_path, f"model{suffix}.pt")

        info = {
            'epoch': self.epoch,
            'best_val_loss': self.best_val_loss,
            'val_loss': val_loss,
            'elektronn3.__version__': elektronn3.__version__,
            'env_info': collect_env.get_pretty_env_info() 
        }

        info = {k: str(v) for k, v in info.items()}

        torch.save({
            'model_state_dict': model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'info': info
        }, state_dict_path)

        log(f'Saved state_dict as {state_dict_path}')
        pts_model_path = f"{model_path}"

        try:
            torch.save(model, model_path)
            log(f'Saved model as {model_path}')
        
        except (TypeError, PicklingError) as exc:
            raise exc

        finally:
            model.training = model_trainmode
        
        if os.path.isfile(pts_model_path):
            with zipfile.ZipFile(pts_model_path, 'a', compression=zipfile.ZIP_DEFLATED) as zfile:
                infostr = pprint.pformat(info, indent=2, width=120)
                zfile.writestr('info.txt', infostr)

    def _log_basic(self, stats, misc):
        tr_loss = stats["tr_loss"][-1]
        val_loss = stats["val_loss"][-1]
        val_acc = stats["val_acc"][-1]
        tr_speed = misc['tr_speed']
        t = pretty_string_time(self._timer.t_passed)
        text = f'epoch={self.epoch:02d}, tr_loss={tr_loss:.3f}, val_loss={val_loss:.3f}, val_acc={val_acc:.2f} '
        text += f'tr_speed={tr_speed:.2f} it/s, {t}'
        logger.info(text)

    def _log_to_tensorboard(
        self,
        stats: Dict,
        misc: Dict,
        file_stats: Optional[Dict] = None
    ) -> None:
        if self.tb:
            try:
                self._tb_log_scalars(stats, "stats")
                self._tb_log_scalars(misc, "misc")
                if file_stats is not None:
                    self._tb_log_scalars(file_stats, "file_stats")
                self._tb_log_histograms()
            except Exception as exc:
                logger.exception('Error occured while logging to tensorboard:')
                raise exc

    def _tb_log_scalars(
            self,
            scalars: Dict[str, float],
            tag: str = 'default'
    ) -> None:
        for key, value in scalars.items():
            if isinstance(value, (list, tuple, np.ndarray)):
                for i in range(len(value)):
                    if not np.isnan(value[i]):
                        self.tb.add_scalar(f'{tag}/{key}', value[i], self.epoch - len(value) + i)
            elif not np.isnan(value):
                self.tb.add_scalar(f'{tag}/{key}', value, self.epoch)

    def _tb_log_histograms(self) -> None:
        """Log histograms of model parameters and their current gradients.

        Make sure to run this between ``backward()`` and ``zero_grad()``,
        because otherwise gradient histograms will only consist of zeros.
        """
        for name, param in self.model.named_parameters():
            self.tb.add_histogram(f'param/{name}', param, self.epoch)
            grad = param.grad if param.grad is not None else torch.tensor(0)
            self.tb.add_histogram(f'grad/{name}', grad, self.epoch)



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany

"""
adopted from https://github.com/pochih/FCN-pytorch/blob/master/python/fcn.py
LICENSE https://github.com/meetshah1995/pytorch-semseg/blob/master/LICENSE

"""

from __future__ import print_function

import torch
import torch.nn as nn
import torch.optim as optim
from torchvision import models
from torchvision.models.vgg import VGG


def resize_conv_ala_distill(in_feat, out_feat, kernel_size, stride, padding,
                                  output_padding, dilation):
    """
    # TODO: needs refinement to work with arbitrary kernel size, stride and padding etc.
    https://distill.pub/2016/deconv-checkerboard/
    https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/issues/190
    Args:
        in_feat ():
        out_feat ():
        kernel_size ():
        stride ():
        padding ():
        output_padding ():
        dilation ():

    Returns:

    """
    return nn.Sequential(nn.UpsamplingNearest2d(scale_factor=stride),
           nn.ReflectionPad2d(padding),
           nn.Conv2d(in_feat, out_feat, kernel_size=kernel_size, stride=1,
           dilation=dilation))


class FCN32s(nn.Module):

    def __init__(self, base_net, n_class):
        super().__init__()
        self.n_class = n_class
        self.base_net = base_net
        self.relu    = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1     = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2     = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3     = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4     = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5     = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, n_class, kernel_size=1)

    def forward(self, x):
        output = self.base_net(x)
        x5 = output['x5']  # size=(N, 512, x.H/32, x.W/32)

        score = self.bn1(self.relu(self.deconv1(x5)))     # size=(N, 512, x.H/16, x.W/16)
        score = self.bn2(self.relu(self.deconv2(score)))  # size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)                    # size=(N, n_class, x.H/1, x.W/1)

        return score  # size=(N, n_class, x.H/1, x.W/1)


class FCN16s(nn.Module):

    def __init__(self, base_net, n_class):
        super().__init__()
        self.n_class = n_class
        self.base_net = base_net
        self.relu    = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1     = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2     = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3     = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4     = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5     = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, n_class, kernel_size=1)

    def forward(self, x):
        output = self.base_net(x)
        x5 = output['x5']  # size=(N, 512, x.H/32, x.W/32)
        x4 = output['x4']  # size=(N, 512, x.H/16, x.W/16)

        score = self.relu(self.deconv1(x5))               # size=(N, 512, x.H/16, x.W/16)
        score = self.bn1(score + x4)                      # element-wise add, size=(N, 512, x.H/16, x.W/16)
        score = self.bn2(self.relu(self.deconv2(score)))  # size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)                    # size=(N, n_class, x.H/1, x.W/1)

        return score  # size=(N, n_class, x.H/1, x.W/1)


class FCN8s(nn.Module):

    def __init__(self, base_net, n_class):
        super().__init__()
        self.n_class = n_class
        self.base_net = base_net
        self.relu    = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1     = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2     = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3     = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4     = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5     = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, n_class, kernel_size=1)

    def forward(self, x):
        output = self.base_net(x)
        x5 = output['x5']  # size=(N, 512, x.H/32, x.W/32)
        x4 = output['x4']  # size=(N, 512, x.H/16, x.W/16)
        x3 = output['x3']  # size=(N, 256, x.H/8,  x.W/8)

        score = self.relu(self.deconv1(x5))               # size=(N, 512, x.H/16, x.W/16)
        score = self.bn1(score + x4)                      # element-wise add, size=(N, 512, x.H/16, x.W/16)
        score = self.relu(self.deconv2(score))            # size=(N, 256, x.H/8, x.W/8)
        score = self.bn2(score + x3)                      # element-wise add, size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)                    # size=(N, n_class, x.H/1, x.W/1)

        return score  # size=(N, n_class, x.H/1, x.W/1)


class FCNs(nn.Module):

    def __init__(self, base_net, n_class):
        super().__init__()
        self.n_class = n_class
        self.base_net = base_net
        self.relu    = nn.ReLU(inplace=True)
        self.deconv1 = nn.ConvTranspose2d(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        # self.deconv1 = resize_conv_ala_distill(512, 512, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn1     = nn.BatchNorm2d(512)
        self.deconv2 = nn.ConvTranspose2d(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        # self.deconv2 = resize_conv_ala_distill(512, 256, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn2     = nn.BatchNorm2d(256)
        self.deconv3 = nn.ConvTranspose2d(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        # self.deconv3 = resize_conv_ala_distill(256, 128, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn3     = nn.BatchNorm2d(128)
        self.deconv4 = nn.ConvTranspose2d(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        # self.deconv4 = resize_conv_ala_distill(128, 64, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn4     = nn.BatchNorm2d(64)
        self.deconv5 = nn.ConvTranspose2d(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        # self.deconv5 = resize_conv_ala_distill(64, 32, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)
        self.bn5     = nn.BatchNorm2d(32)
        self.classifier = nn.Conv2d(32, n_class, kernel_size=1)

    def forward(self, x):
        output = self.base_net(x)
        x5 = output['x5']  # size=(N, 512, x.H/32, x.W/32)
        x4 = output['x4']  # size=(N, 512, x.H/16, x.W/16)
        x3 = output['x3']  # size=(N, 256, x.H/8,  x.W/8)
        x2 = output['x2']  # size=(N, 128, x.H/4,  x.W/4)
        x1 = output['x1']  # size=(N, 64, x.H/2,  x.W/2)

        score = self.bn1(self.relu(self.deconv1(x5)))     # size=(N, 512, x.H/16, x.W/16)
        score = score + x4                                # element-wise add, size=(N, 512, x.H/16, x.W/16)
        score = self.bn2(self.relu(self.deconv2(score)))  # size=(N, 256, x.H/8, x.W/8)
        score = score + x3                                # element-wise add, size=(N, 256, x.H/8, x.W/8)
        score = self.bn3(self.relu(self.deconv3(score)))  # size=(N, 128, x.H/4, x.W/4)
        score = score + x2                                # element-wise add, size=(N, 128, x.H/4, x.W/4)
        score = self.bn4(self.relu(self.deconv4(score)))  # size=(N, 64, x.H/2, x.W/2)
        score = score + x1                                # element-wise add, size=(N, 64, x.H/2, x.W/2)
        score = self.bn5(self.relu(self.deconv5(score)))  # size=(N, 32, x.H, x.W)
        score = self.classifier(score)                    # size=(N, n_class, x.H/1, x.W/1)

        return score  # size=(N, n_class, x.H/1, x.W/1)


class VGGNet(VGG):
    def __init__(self, model='vgg16', requires_grad=True,
                 remove_fc=True, show_params=False, in_channels=3, batch_norm=False):
        super().__init__(make_layers(cfg[model], batch_norm, in_channels))
        self.ranges = ranges[model]

        if not requires_grad:
            for param in super().parameters():
                param.requires_grad = False

        if remove_fc:  # delete redundant fully-connected layer params, can save memory
            del self.classifier

        if show_params:
            for name, param in self.named_parameters():
                print(name, param.size())

    def forward(self, x):
        output = {}

        # get the output of each maxpooling layer (5 maxpool in VGG net)
        for idx in range(len(self.ranges)):
            for layer in range(self.ranges[idx][0], self.ranges[idx][1]):
                x = self.features[layer](x)
            output["x%d"%(idx+1)] = x

        return output


ranges = {
    'vgg11': ((0, 3), (3, 6),  (6, 11),  (11, 16), (16, 21)),
    'vgg13': ((0, 5), (5, 10), (10, 15), (15, 20), (20, 25)),
    'vgg16': ((0, 5), (5, 10), (10, 17), (17, 24), (24, 31)),
    'vgg19': ((0, 5), (5, 10), (10, 19), (19, 28), (28, 37))
}

# cropped version from https://github.com/pytorch/vision/blob/master/torchvision/models/vgg.py
cfg = {
    'vgg11': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'vgg13': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'vgg16': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'vgg19': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}

def make_layers(cfg, batch_norm=False, in_channels=3):
    layers = []
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
            else:
                layers += [conv2d, nn.ReLU(inplace=True)]
            in_channels = v
    return nn.Sequential(*layers)